/**
 ****************************************************************************
 *
 *  @file   pktcache.cpp
 *  @author Ole-Ivar Holthe
 *  @date   2007-02-27
 *  @brief  Implements the CGSPktCache class.
 *
 *  Copyright � Gridmedia Technologies 1996-2008. All Rights Reserved.
 *
 ****************************************************************************/

#ifdef _WIN32
#include "internal.h"
#else // LINUX_COMPILE
#include "libgxsys.h"
#endif
#include "pktcache.h"

#ifndef GSTRACE
	void GSTRACE2(const char* fmt, ...) {}
	#define GSTRACE GSTRACE2
#endif // GSTRACE

/**
 ****************************************************************************
 * 
 *  @class  CGSPktCache
 *  @brief  The packet cache class.
 *
 ****************************************************************************/

CGSPktCache::CGSPktCache()
{
    memset(m_pkts, 0, sizeof(GXPKT)*MAXPKTS);
    memset(&m_ret, 0, sizeof(GXPKT));
    m_i = 0;

    m_nPktSeqNoLastPut = 0; // NEW!
    m_nPktCountPut = 0; // NEW!
}

CGSPktCache::~CGSPktCache()
{
    Clear();
}


/*
 ****************************************************************************
 * 
 *  METHODS
 *
 ****************************************************************************/

void CGSPktCache::Clear()
{
    for (int i = 0; i < MAXPKTS; i++) {
        if (m_pkts[i].pPktData != NULL)
            free(m_pkts[i].pPktData);
    }
    if (m_ret.pPktData != NULL)
        free(m_ret.pPktData);
    m_i = 0;
    memset(m_pkts, 0, sizeof(GXPKT)*MAXPKTS);
    memset(&m_ret, 0, sizeof(GXPKT));

    m_nPktSeqNoLastPut = 0; // NEW!
    m_nPktCountPut = 0; // NEW!
}

BOOL CGSPktCache::Put(BYTE nType, LONGLONG nTimestamp, int nFlags, 
        int nPktNum, int nPktCount, UINT nPktSeqNo, BYTE* pData, int nBytes)
{
    // detect: cache overflow
    if (m_pkts[m_i].nType != 0) {
        GSTRACE("CGSPktCache::Put(): Cache overflow!\n");
    }

    // set attribs
    m_pkts[m_i].nType = nType;
    m_pkts[m_i].nTimestamp = nTimestamp;
    m_pkts[m_i].nFlags = nFlags;
    m_pkts[m_i].nPktNum = nPktNum;
    m_pkts[m_i].nPktCount = nPktCount;
    m_pkts[m_i].nPktSeqNo = nPktSeqNo;

    // alloc or realloc
    if (m_pkts[m_i].pPktData == NULL)
        m_pkts[m_i].pPktData = (BYTE*)malloc(nBytes);
#ifdef LINUX_ONLY
	else if (malloc_usable_size(m_pkts[m_i].pPktData) < (size_t)nBytes)
#else // MAC
    else if (malloc_size(m_pkts[m_i].pPktData) < (size_t)nBytes)
#endif
        m_pkts[m_i].pPktData = (BYTE*)realloc(m_pkts[m_i].pPktData, nBytes);

    // out of memory?
    if (m_pkts[m_i].pPktData == NULL)
        return FALSE;

    // copy data
    memcpy(m_pkts[m_i].pPktData, pData, nBytes);
    m_pkts[m_i].nPktBytes = nBytes;

    // analysis
    m_nPktSeqNoLastPut = nPktSeqNo;
    m_nPktCountPut++;

    // inc index
    m_i++;
    if (m_i >= MAXPKTS)
        m_i = 0;

    return TRUE;
}

CGSPktCache::GXPKT* CGSPktCache::Get()
{
    // find minimum timestamp
    int imin = -1;
    int i;
    LONGLONG tmin = MAXLONGLONG;
    for (i = 0; i < MAXPKTS; i++) {
        if ((m_pkts[i].nType != 0) && (m_pkts[i].nTimestamp < tmin)) {
            tmin = m_pkts[i].nTimestamp;
            imin = i;
        }
    }
    if (tmin == MAXLONGLONG)
        return NULL; // cache is completely empty!


    // check if we have all pkts
    int nType = m_pkts[imin].nType;
    int nPktCount = 0;
    int nPktBytes = 0;
    int nPktCountNewer = 0;
    for (i = 0; i < MAXPKTS; i++) {
        if ((m_pkts[i].nType == nType) && (m_pkts[i].nTimestamp == tmin)) {
            nPktCount++;
            nPktBytes += m_pkts[i].nPktBytes;
        }
        if (m_pkts[i].nTimestamp > tmin) {
            nPktCountNewer++;
        }
    }


    // we don't have all pkts yet
    if ((nPktCount < m_pkts[imin].nPktCount) && (nPktCountNewer < 2 /*?*/))
        return NULL;
    // Detect: Incomplete a/v-pkt (i.e. fragmented and has atleast one lost fragment)
    if (nPktCount < m_pkts[imin].nPktCount) {
        GSTRACE("CGSPktCache::Get(): Incomplete vid-pkt with %d of %d pkts\n", nPktCount, 			m_pkts[imin].nPktCount);
    }
    nPktCount = m_pkts[imin].nPktCount; // for safety!


    // assemble (what we have)
    // -----------------------

    // set attribs
    m_ret.nType = m_pkts[imin].nType;
    m_ret.nTimestamp = m_pkts[imin].nTimestamp;
    m_ret.nFlags = m_pkts[imin].nFlags;
    m_ret.nPktNum = m_pkts[imin].nPktNum;
    m_ret.nPktCount = m_pkts[imin].nPktCount;
    m_ret.nPktSeqNo = m_pkts[imin].nPktSeqNo;

    // alloc or realloc
    if (m_ret.pPktData == NULL)
        m_ret.pPktData = (BYTE*)malloc(nPktBytes);
#ifdef LINUX_ONLY
	else if (malloc_usable_size(m_ret.pPktData) < (size_t)nPktBytes)
#else // MAC
    else if (malloc_size(m_ret.pPktData) < (size_t)nPktBytes)
#endif
        m_ret.pPktData = (BYTE*)realloc(m_ret.pPktData, nPktBytes);

    // out of memory?
    if (m_ret.pPktData == NULL)
        return NULL;

    // clear memory
    memset(m_ret.pPktData, 0, nPktBytes);
    m_ret.nPktBytes = nPktBytes;

    // copy data
    int j;
    BYTE* p = m_ret.pPktData;
    for (i = 0; i < nPktCount; i++) {
        for (j = 0; j < MAXPKTS; j++) {
            if ((m_pkts[j].nType == nType) && (m_pkts[j].nTimestamp == tmin) && (m_pkts[j].nPktNum == i)) {
                // copy pkt data
                memcpy(p, m_pkts[j].pPktData, m_pkts[j].nPktBytes);
                // next line
                p += m_pkts[j].nPktBytes;
                // clear pkt
                m_pkts[j].nType = 0;
                m_pkts[j].nTimestamp = 0;
            }
        }
    }

    return &m_ret;
}
