/**
 ****************************************************************************
 *
 *  @file   sockudp.cpp
 *  @author Ole-Ivar Holthe
 *  @date   2007-02-22
 *  @brief  Implements the CGSSockUDP class.
 *
 *  Copyright  Gridmedia Technologies 1996-2008. All Rights Reserved.
 *
 ****************************************************************************/

#ifdef _WIN32
#include "internal.h"
#else // LINUX_COMPILE
#include "libgxsys.h"
#endif
#include "sockudp.h"

/**
 ****************************************************************************
 * 
 *  @class  CGSSockUDP
 *  @brief  The UDP socket class.
 *
 ****************************************************************************/

CGSSockUDP::CGSSockUDP()
{
    m_sd = (SOCKET)NULL;
    memset(&m_addrBind, 0, sizeof(struct sockaddr_in));
}

CGSSockUDP::~CGSSockUDP()
{
    Destroy();
}


/*
 ****************************************************************************
 * 
 *  METHODS
 *
 ****************************************************************************/

BOOL CGSSockUDP::Create()
{
    // Create a datagram socket
	m_sd = socket(AF_INET, SOCK_DGRAM, IPPROTO_UDP);
	return (m_sd != INVALID_SOCKET);
}

void CGSSockUDP::Destroy()
{
    // Close socket
#ifdef LINUX_COMPILE
    if (m_sd != (SOCKET) NULL)
	    close(m_sd);
    m_sd = (SOCKET) NULL;
#else
    if (m_sd != NULL)
        closesocket(m_sd);
    m_sd = NULL;
#endif // LINUX_COMPILE

}

BOOL CGSSockUDP::Bind(LPCTSTR szIP, int nPort)
{
    // Clear m_addrBind
	memset(&m_addrBind, 0, sizeof(struct sockaddr_in));

	// Set family and port
	m_addrBind.sin_family = AF_INET;
	m_addrBind.sin_port = htons(nPort);

	// Set IP address
	if (szIP != NULL)
    {
        // Read IP address
		m_addrBind.sin_addr.s_addr = inet_addr(szIP);
        if (m_addrBind.sin_addr.s_addr == INADDR_NONE)
            return FALSE; // Typo!
    }
    else
    {
		// Get host name of this computer
        char host_name[256];
        struct hostent* hp;
		gethostname(host_name, sizeof(host_name));
		hp = gethostbyname(host_name);
		if (hp == NULL)
			return FALSE; // Couldn't get host name!
        
        //LINUX_COMPILE
        //NOTE OIH 2008-04-23: Sorry, not thread safe.
		//m_addrBind.sin_addr.s_addr = inet_addr(inet_ntoa(*( struct in_addr*)( hp -> h_addr_list[0])));
	// ADDED OM 2008-04-28: Thread safe alternative
	#ifdef LINUX_COMPILE
		m_addrBind.sin_addr.s_addr = (( struct in_addr*)( hp -> h_addr_list[0]))->s_addr;
		//inet_aton( inet_ntoa(*( struct in_addr*)( hp -> h_addr_list[0]))), 
	#else  
		// Assign the address
		m_addrBind.sin_addr.S_un.S_un_b.s_b1 = hp->h_addr_list[0][0];
		m_addrBind.sin_addr.S_un.S_un_b.s_b2 = hp->h_addr_list[0][1];
		m_addrBind.sin_addr.S_un.S_un_b.s_b3 = hp->h_addr_list[0][2];
		m_addrBind.sin_addr.S_un.S_un_b.s_b4 = hp->h_addr_list[0][3];
	#endif // LINUX_COMPILE
	}
	
	// Bind address to socket
	if (bind(m_sd, (struct sockaddr*)&m_addrBind, sizeof(struct sockaddr_in)) == -1)
        return FALSE; // Could not bind name to socket!

    return TRUE;
}


/*
 ****************************************************************************
 * 
 *  GLOBALS
 *
 ****************************************************************************/

/**
 *  Resolves the IP or DNS address.
 */
BOOL CGSSockUDP::Resolve(struct sockaddr_in* saddr, LPCTSTR szAddress, int port)
{
    // Clear saddr
	memset(saddr, 0, sizeof(struct sockaddr_in));

	// Set family and port
	saddr->sin_family = AF_INET;
	saddr->sin_port = htons(port);

    // Resolve IP address
    ULONG addr = inet_addr(szAddress);
    if (addr == INADDR_NONE)
    {
        // Attempt DNS resolving
#ifdef LINUX_COMPILE
        struct hostent* res;
#else
        HOSTENT* res;
#endif
        res = gethostbyname(szAddress);
        if (res != NULL)
            addr = (*reinterpret_cast<ULONG*>(*res->h_addr_list));
    }
    if (addr == INADDR_NONE)
        return FALSE;

    // Set address
#ifdef LINUX_COMPILE
    saddr->sin_addr.s_addr = addr;
#else
    saddr->sin_addr.S_un.S_addr = addr;
#endif
    return TRUE;
}


/*
 ****************************************************************************
 * 
 *  WIN32
 *
 ****************************************************************************/

#ifdef _WIN32

/**
 *  Initialize use of Windows Sockets v2.0.
 *  WSACleanup() must be called for cleanup.
 */
BOOL CGSSockUDP::SockLibInit(WSADATA* lpwsaData)
{
    // Initialize Winsock library
    WSADATA wsaData;
    if (lpwsaData == NULL)
        lpwsaData = &wsaData;

    WORD wVersionRequested = MAKEWORD(2, 0);
    int nResult = WSAStartup(wVersionRequested, lpwsaData);
    if (nResult != 0)
        return FALSE;

    if (LOBYTE(lpwsaData->wVersion) != 2 || HIBYTE(lpwsaData->wVersion) != 0) {
        WSACleanup();
        return FALSE;
    }

    return TRUE;
}

#endif // _WIN32
