/**
 * @file		GXLinuxClient.h
 * @author 		Ola Mogstad (olamo@stud.ntnu.no)
 * @project		Geelix LiveGames Linux
 * @updated		23/03/08
 * @brief		Application behavior header file.
 */

#ifndef GXLINUXCLIENT_H
#define GXLINUXCLIENT_H

class GXLinuxClient : CGXClientLGV
{
public:
	// constructors & destructors
	GXLinuxClient();
	~GXLinuxClient();
	
	// functions
	int InitClient(UserInfo userInfo);
	void CloseClient(const char* reason);
	void DoWork();
	void DoWorkOnFile();
	int DoConnect();
	int SetupFilePlay(const char* url);
	int SetupDecRend();
	int DoDecQueue(CGSPktCache::GXPKT* pPktCC);
	attribute_deprecated int DoConnectInputTest();
	void Stop() { running = 0; }
	BOOL OnEchoBack();

	// setters & getters
#ifdef LINUX_ONLY
	void SetGuiComponents(Gtk::Statusbar* stb, Gtk::Button* btn)	{ GUIstatusbar = stb; GUIconnectbutton = btn; }
#endif
	int GetConnected()												{ return ( (state == CONNECTED) || (state == FILEPLAY) ); }
	
private:
	// member functions
	void SetUserInfo(UserInfo userInfo)		{ ui = userInfo; ui.infoRetrieved = 1; }
	int ReceiveVideoPacket(int flags);
	void SetStatusbar(const char* msg)		{ 
			#ifdef LINUX_ONLY 
				if (GUIstatusbar != NULL) 
					GUIstatusbar->push(msg,0); 
			#endif 
			LOG.MLog("Statusbar: %s",msg); 
		}
	void ToggleConnectButton()				{
			#ifdef LINUX_ONLY 
				if ( state != IDLE ) 
					GUIconnectbutton->set_label("Disconnect"); 
				else if ( state == IDLE ) 
					GUIconnectbutton->set_label("Connect"); 
			#endif 
		}
	void SetFrequency(int f) 				{ frequency = f; }
	void SetConnected() 					{ PktCache.Clear(); SetStatusbar("Connected"); ToggleConnectButton(); state = CONNECTED; }
	
	// deprecated
	attribute_deprecated int ReadStreamLoop();
	attribute_deprecated int DoConnectLiveStart();
	attribute_deprecated int DoConnectHUD();
	attribute_deprecated void PrintStats();
	
	// data members
	int running;						// Contols if the loop is running or not
	int frequency;						// Controls loop speed
	UserInfo 			ui;				// Holds all user information from GUI
	SocketInfo 			si;				// Holds all socket information
	HUDStreamInfo 		hsi;			// Holds all streaming information
#ifdef LINUX_ONLY
	Gtk::Statusbar*		GUIstatusbar;	// A statusbar that information can be relayed to
	Gtk::Button*		GUIconnectbutton;
#endif
	GXLinuxDecoder 		gxl_decoder;	// The decoder
	GXLinuxRenderer 	gxl_renderer;	// The renderer (check if enabled)
	GXLinuxInputHandler gxl_input;		// The input handler
	CGSPktCache 		PktCache;		// The packet cache for assembling multiple packets
	
	char* packetBuf;
	int decodedFrameSize;
	AVFrame* bfrFrame;
	
	ifstream file;						// File for reading .pk5 files
	
#ifdef PERFORMANCETEST
	// Performance testing
	StopWatch SW_Total;
	StopWatch SW_Decoder;
	StopWatch SW_Active;
#endif	

};

/**
 * @function	Default constructor.
 */
GXLinuxClient::GXLinuxClient()
{
	// Default values
	frequency = 100;
	
#ifdef USEGXSYS
	LOG.MLog("USEGXSYS: enabled (only option)");
#else
	LOG.MLog("ERROR: USEGXSYS was not defined");
	exit(1);
#endif
	
#ifdef USESWSCALE
	LOG.MLog("USESWSCALE: enabled (will use libswscale and not img_convert)");
#else
	LOG.MLog("USESWSCALE: disabled");
#endif
	
#ifdef USEFFMPEGLIB
	LOG.MLog("ERROR: USEFFMPEGLIB was defined but is deprecated!");
	exit(1);
#else
	LOG.MLog("USEFFMPEGLIB: disabled");
#endif
	
	if (RENDER_AUDIO)
		LOG.MLog("RENDER_AUDIO: enabled (will play sound)");
	else
		LOG.MLog("RENDER_AUDIO: disabled (no sound)");
	
	if (RENDER_VIDEO)
		LOG.MLog("RENDER_VIDEO: enabled (will show frames)");
	else
		LOG.MLog("RENDER_VIDEO: disabled (no frames)");
	
}

/**
 * @function	Default destructor.
 */
GXLinuxClient::~GXLinuxClient()
{
	// TODO: FIX!
	//free(packetBuf);
	//free(frameBuf);
	if (GetConnected())
		CloseClient("Unexpected termination");
}

/**
 * @function	InitClient()
 * @brief		Sets up the socket information.
 * @returns		-1 on error, 0 if ok
 * @note		Does not init decoder or renderer since parameters must be retrieved first
 * 					which requires connection.
 */
int GXLinuxClient::InitClient(UserInfo userInfo)
{
	SetUserInfo(userInfo);
	
	packetBuf = (char*) calloc(MAX_DATA_PACKET,sizeof(char));
	decodedFrameSize = 0;
	
	if ( packetBuf == NULL )
	{
		LOG.MLog("Failed to allocate packet/frame cache");
		return -1;
	}
	
#ifdef PERFORMANCETEST
	SW_Total.LapStart();
#endif
    
    return 0;
}

/**
 * @function	CloseClient()
 * @brief		Should disconnect the client from everything
 */
void GXLinuxClient::CloseClient(const char* reason)
{
	if (!GetConnected())
		SetStatusbar("Client already closed - will not try to notify server");
	else
	{
		if (DoSendBye())
			LOG.MLog("Successfully sent BYE packet to server");
		else
			LOG.MLog("Failed to send QUIT packet to server");
	}

	ToggleConnectButton();
	
	// Closing sockets
	m_sock.Destroy();

	// Closing the other modules
	gxl_renderer.CloseRenderer();
	gxl_decoder.CloseDecoder();	
	gxl_input.CloseInputHandler();

	if (file.is_open())
		file.close();

	char tmp[256];
	sprintf(tmp,"Disconnected: %s",reason);
	SetStatusbar(tmp);
	
	state = IDLE;
	ToggleConnectButton();
	
#ifdef PERFORMANCETEST
	SW_Decoder.LapStop();
	printf("SW_Decoder: "); SW_Decoder.Print(); SW_Decoder.PrintData();
	SW_Total.LapStop();
	printf("SW_Total: "); SW_Total.Print();
	SW_Active.LapStop();
	printf("SW_Active: "); SW_Active.Print();
#endif
}

/**
 * @function	StartThreadInputLoop()
 * @brief		Necessary for SDL threads to work as SDL uses old C semantics.?
 */
/*int StartThreadInputLoop(void *obj)
{
	GXLinuxInputHandler* inputhandler = static_cast<GXLinuxInputHandler*>(obj);
	return inputhandler->DoInputLoop();
}*/

/**
 * @function	DoWork()
 * @brief		One iteration of the main loop. Calls all DoWorks
 */
void GXLinuxClient::DoWork()
{
	if (state == IDLE)			// do nothing
		return;	
	else if (state == WAIT_LS)	// deprecated..?
		return;
	else if (state == WAIT_HUD) // check for response
	{
		if (DoListen() > 0)	// check for data on socket
		{
			if (OnConnect() == FALSE)
			{
				char closemsg[256];
				if (strlen( (char*)m_pktbuf ) > 0)
					sprintf(closemsg,"Failed to connect to server (%s)",(char*)m_pktbuf);
				else
					sprintf(closemsg,"Failed to connect to server");
				CloseClient(closemsg);
			}
			else
			{
				SetConnected();
				SetupDecRend();
				
				// Send stuff for testing
				char logstat[10000];
				sprintf(logstat,"LOG\n%s",SYSINFO_LINUX());
				DoSendMsg(logstat);
			}
		}
		return;
	}
	else if (state == FILEPLAY)
	{
		DoWorkOnFile();
		return;
	}
	//else if (state == CONNECTED)
#ifdef PERFORMANCETEST
	SW_Active.LapStart();
#endif
	
	// Let input handler do work
	if (gxl_input.DoWork() == -1)
		CloseClient("Requested by user");
	
	// Get packet from cache
	CGSPktCache::GXPKT* pPktCC = m_cache.Get();
	
	// If there was a packet pending, decode and queue it in renderer
	if (pPktCC != NULL)
		DoDecQueue(pPktCC);
	
	// Then, render items from queues
	gxl_renderer.DoWork();

	// Check for more packets and put potential packets in packet cache
	if (DoListen() > 0)
		OnVidPkt();
	
#ifdef PERFORMANCETEST
	SW_Active.LapStop();
#endif
	
}

unsigned long nTmpLoss = 0;
unsigned long nTmpPackets = 0;

/**
 * @function	OnEchoBack()
 * @brief		Redefines virtual function from CGXClient
 */
BOOL GXLinuxClient::OnEchoBack()
{
	// Set title bar
	char newtitlebar[256];
	double lossPct = 100.0*((double)(m_cache.m_nPktSeqNoLastPut-m_cache.m_nPktCountPut+1) - nTmpLoss) / (double)(m_cache.m_nPktCountPut - nTmpPackets);
	if ( (lossPct <= 100.0) && (lossPct >= 0.0) )
		sprintf(newtitlebar,"LiveGames Video   (Ping: %4lu ms   Loss: %5.2f \%)", m_nEchoLatency, lossPct);
	else
		sprintf(newtitlebar,"LiveGames Video   (Ping: %4lu ms   Loss: -.-- \%)", m_nEchoLatency);
	gxl_renderer.SetTitlebar(newtitlebar);
	nTmpLoss = m_cache.m_nPktSeqNoLastPut-m_cache.m_nPktCountPut+1;
	nTmpPackets = m_cache.m_nPktCountPut;

	
	// Send LOG msg to server
	char logmsg[256];
	sprintf(logmsg,"LOG:Roundtrip latency: %lums   Packet loss: %lu   Packets: %lu\n",
			m_nEchoLatency, m_cache.m_nPktSeqNoLastPut-m_cache.m_nPktCountPut+1, m_cache.m_nPktCountPut);
	DoSendMsg(logmsg);

	return TRUE;
}

/**
 * @function	DoConnect()
 * @brief		Goes through the steps of setting up a connection to a HUD. 
 * @returns		-1 if error, 0 if ok
 */
int GXLinuxClient::DoConnect()
{	
	LOG.MLog("Running DoConnect() on ui.server = %s",ui.server);
	
	SetStatusbar("Connecting to server ...");
	
	if (strncmp(ui.game,"(Running",7) == 0)
		LOG.MLog("Connecting to running game, skipping DoConnectLiveStart()");
	else
	{
		LOG.MLog("Running game was not chosen, but forcing it");
		SetStatusbar("Forcing a connect to (Running game)");
	}
	
	// Setting video parameters that will be requested
    strcpy(m_szLoginName, ui.username);
    strcpy(m_szLoginPassword, ui.password);
    m_nLoginGameID = -1;
    m_nPacketSize = ui.packetsize;
    
    int tmp_bitrate = 0;
    sscanf(ui.profile,"%i",&tmp_bitrate);
    m_nVidBitrate = tmp_bitrate * 1000;	// Multiply with 1000 since GUI uses Kbps and protocol bps
    m_nVidFps = ui.framerate;
	m_nVidFmtDX = PIX_FMT_YUV420P;
	m_nVidCodec = 19;						// We request WMV2 for now
	m_nVidWidth = 0;						// Server ignores this
	m_nVidHeight = 0;
	
	if ( ui.size[0] == 'N' )
		m_nVidWidth = 0;
	else if ( ui.size[0] == 'H' )
		m_nVidWidth = -1;
	else if ( ui.size[0] == 'Q' )
		m_nVidWidth = -2;
	
	// Setting audio parameters that will be requested
	if ( ui.abitrate[0] == '6' )
		m_nAudBitrate = 64000;
	else if ( ui.abitrate[0] == '1' )
			m_nAudBitrate = 128000;
	else if ( ui.abitrate[0] == 'N' )
			m_nAudBitrate = 0;
	
    m_nAudCodec = CODEC_ID_WMAV2;
    m_nAudChannels = 0;		// ignored in request
    m_nAudSampleRate = 0;		// ignored in request
    m_nAudBitsPerSample = 0;	// ignored in request
	
	// Check if server address is of file:// format
	if (strncmp(ui.server,"file:",5) == 0)
	{
		return SetupFilePlay(ui.server+7);
	}
    
	if ( DoSendConnectVideo(ui.server, ui.port) == FALSE )
	{
		this->CloseClient("Connecting to game HUD failed");
		return -1;
	}
	
	state = WAIT_HUD;
	ToggleConnectButton();
	
	return 0;
}

/**
 * @function	SetupDecRen()
 * @brief		Sets up decoders and renderers
 */
int GXLinuxClient::SetupDecRend()
{	
	if (gxl_decoder.InitVideoDecoder(*reinterpret_cast<PixelFormat*>(&m_nVidFmtDX),
				*reinterpret_cast<CodecID*>(&m_nVidCodec),
				m_nVidWidth,
				m_nVidHeight,
				PIX_FMT_YUV420P) == -1)
	{
		LOG.MLog("ERROR: Failed to init the video decoder");
		this->CloseClient("Initializing the video decoder failed");
		return -1;
	}
	
	if (gxl_decoder.InitAudioDecoder(static_cast<enum CodecID>(m_nAudCodec), m_nAudChannels, m_nAudSampleRate) == -1)
	{
		LOG.MLog("WARNING: Failed to init the audio decoder - continuing without it (set RENDER_AUDIO = 0)");
		RENDER_AUDIO = 0;
	}

	// Initiate the renderer
	if (gxl_renderer.InitVideoRenderer(m_nVidWidth,m_nVidHeight,gxl_decoder.GetVideoOutFormat()) == -1)
	{
		LOG.MLog("ERROR: Failed to init the video renderer");
		this->CloseClient("Initializing the video renderer failed");
		return -1;
	}

	if (ui.fullscreen == 1)
		gxl_renderer.ToggleFullscreen();
	
	if (ui.grabinput == 1)
		gxl_renderer.SetGrabInput(1);
	
	if (gxl_renderer.InitAudioRenderer(m_nAudSampleRate,m_nAudChannels, m_nAudBitsPerSample) == -1)
	{
		LOG.MLog("WARNING: Failed to init the audio renderer - continuing without it (set RENDER_AUDIO = 0)");
		RENDER_AUDIO = 0;
	}
	
	if (gxl_input.InitInputHandlerGXSYS(ui.server,ui.port+2,&gxl_renderer,this) == -1)
	{
		LOG.MLog("ERROR: Failed to init the input handler");
		this->CloseClient("Initializing the input handler failed");
		return -1;
	}

	// Initiate the buffer frame
	bfrFrame = avcodec_alloc_frame();
	if(bfrFrame == NULL)
	{
		LOG.MLog("ERROR: Failed to create the buffer frame");
		this->CloseClient("Failed to create a buffer frame");
		return -1;
	}
	int numBytes;
	numBytes = avpicture_get_size(gxl_decoder.GetVideoOutFormat(), m_nVidWidth, m_nVidHeight);
	uint8_t* buffer=(uint8_t *)av_malloc(numBytes*sizeof(uint8_t));
	avpicture_fill((AVPicture *)bfrFrame, buffer, gxl_decoder.GetVideoOutFormat(), m_nVidWidth, m_nVidHeight);
	
	return 0;
}

/**
 * @function	DoDecQueue()
 * @brief		Decode a packet and enqueue the output
 * @returns		-1 if it didn't render, 0 if ok
 */
int GXLinuxClient::DoDecQueue(CGSPktCache::GXPKT* pPktCC)
{
	// *** VIDEO FRAMES ***
    if ( (pPktCC->nType == 'V') || (pPktCC->nType == 'v') )
    {
    	
    	if (RENDER_VIDEO == 0)
    		return -1;

        bzero(bfrFrame,sizeof(bfrFrame));
        
#ifdef PERFORMANCETEST
        SW_Decoder.LapStart();
#endif
        int tmp = gxl_decoder.DecodeVideoPacket(pPktCC->pPktData, pPktCC->nPktBytes, bfrFrame, &decodedFrameSize);
#ifdef PERFORMANCETEST
        SW_Decoder.LapStop();
#endif
        
		if (tmp == -1)
		{
			LOG.MLog("WARNING: Failed to decode the frame data. Frame will not be rendered");
			return -1;
		}

		gxl_renderer.EnqueueFrame(bfrFrame);
    }
    
	// *** AUDIO SAMPLES ***
	else if  ( (pPktCC->nType == 'A') || (pPktCC->nType == 'a') )
	{
		if (RENDER_AUDIO == 0)
			return -1;
		
		LOG.DLog("Read audio packet of size=%i",pPktCC->nPktBytes);
		
		// Decode audio samples
		AVPacket pkt;
        char* aBuf = NULL;
        int aBufLen = 0;
        pkt.data = (uint8_t*)av_malloc(pPktCC->nPktBytes);
        memcpy(pkt.data, pPktCC->pPktData, pPktCC->nPktBytes);
        pkt.size = pPktCC->nPktBytes;
            
        if(gxl_decoder.DecodeAudioPacket(&pkt, &aBuf, &aBufLen) == 1)
        	gxl_renderer.EnqueueSamples(aBuf, aBufLen);
        
        av_free(pkt.data);	// Hmm..?
	}
	    
	// *** NOT AUDIO OR VIDEO ***
	else
	{
		LOG.DLog("WARNING: Ignoring packet of type: %c",pPktCC->nType);
		return -1;
	}
    
    return 0;
}

/**
 * @function	SetupFilePlay()
 * @brief		Instead of connecting to a server, reads a stream from a .pk5 file
 * @param		url			Path to .pk5 file with filename
 * @returns		-1 if error, 0 if ok
 */
int GXLinuxClient::SetupFilePlay(const char* url)
{
	SetStatusbar("Opening file");
	LOG.MLog("Running SetupFilePlay on %s",url);
	
	// Sanity checks
	if (strncmp(url+(strlen(url)-4),".pk5",3) != 0)
	{
		SetStatusbar("ERROR: File url is not of .pk5 format");
		return -1;
	}
	
	// Open file
	file.open(url, ios::in | ios::binary);
	if ( file.is_open() == 0 ) 
	{
		SetStatusbar("ERROR: File could not be opened");
		return -1;
	}
	
	// Read header
	BOOL b = TRUE;
	char sign[8];
	char ctmp[128];
	bzero(ctmp,sizeof(ctmp));
	
	LOG.MLog("sizeofs: int=%i ulong=%i sockaddr_in=%i ctmp=%i longlong=%i",
			sizeof(int),sizeof(unsigned long),sizeof(struct sockaddr_in),sizeof(ctmp),sizeof(long long));
	
	b &= ( file.read(sign,8) != NULL );	LOG.MLog("File signature: %s",sign);				// signature(8)
	
	if (strncmp(sign,"PK5000",6) != 0)
		CloseClient("Could not read .pk5 file - invalid format");
	
	b &= ( file.read(ctmp,40) != NULL ); // bzero(ctmp,sizeof(ctmp));								// server-video-ip(40)
	b &= ( file.read(ctmp,sizeof(int)) != NULL ); bzero(ctmp,sizeof(ctmp));							// server-video-port(4)
	b &= ( file.read(m_szLoginName,40) != NULL );	 										// login-name(40)
	b &= ( file.read(reinterpret_cast<char*>(&m_nLoginGameID),sizeof(int)) != NULL );		// login-game-id(4)
	b &= ( file.read(ctmp,sizeof(int)) != NULL );													// login-time-clock(4)
	b &= ( file.read(ctmp,sizeof(struct sockaddr_in)) != NULL );									// client-ip(16)
	
	b &= ( file.read(ctmp,sizeof(unsigned long)) != NULL );				// reserved1(4)
	b &= ( file.read(ctmp,sizeof(unsigned long)) != NULL );				// reserved2(4)
	b &= ( file.read(ctmp,sizeof(unsigned long)) != NULL );				// reserved3(4)
	b &= ( file.read(ctmp,sizeof(unsigned long)) != NULL );				// reserved4(4)
	
	b &= ( file.read(reinterpret_cast<char*>(&m_nPacketSize),sizeof(int)) != NULL );		// packet-size(4)
	b &= ( file.read(reinterpret_cast<char*>(&m_nVidCodec),sizeof(int)) != NULL );		// vid-codec(4)
	b &= ( file.read(reinterpret_cast<char*>(&m_nVidBitrate),sizeof(int)) != NULL );		// vid-bitrate(4)
	b &= ( file.read(reinterpret_cast<char*>(&m_nVidFmtDX),sizeof(int)) != NULL );		// vid-fmtdx(4)
	b &= ( file.read(reinterpret_cast<char*>(&m_nVidWidth),sizeof(int)) != NULL );		// vid-width(4)	
	b &= ( file.read(reinterpret_cast<char*>(&m_nVidHeight),sizeof(int)) != NULL );		// vid-height(4)
	b &= ( file.read(reinterpret_cast<char*>(&m_nVidFps),sizeof(int)) != NULL );			// vid-fps(4)
	b &= ( file.read(reinterpret_cast<char*>(&m_nAudCodec),sizeof(int)) != NULL );		// aud-codec(4)
	b &= ( file.read(reinterpret_cast<char*>(&m_nAudBitrate),sizeof(int)) != NULL );		// aud-bitrate(4)
	b &= ( file.read(reinterpret_cast<char*>(&m_nAudChannels),sizeof(int)) != NULL );		// aud-channels(4)
	b &= ( file.read(reinterpret_cast<char*>(&m_nAudSampleRate),sizeof(int)) != NULL );	// aud-samplerate(4)
	b &= ( file.read(reinterpret_cast<char*>(&m_nAudBitsPerSample),sizeof(int)) != NULL );// aud-bitspersample(4)
	b &= ( file.read(reinterpret_cast<char*>(&m_nInpKBLayout),sizeof(int)) != NULL );		// inp-kblayout(4)
			
	b &= ( file.read(ctmp,sizeof(unsigned long)) != NULL );				// reserved1(4)
	b &= ( file.read(ctmp,sizeof(unsigned long)) != NULL );				// reserved2(4)
	b &= ( file.read(ctmp,sizeof(unsigned long)) != NULL );				// reserved3(4)
	b &= ( file.read(ctmp,sizeof(unsigned long)) != NULL );				// reserved4(4)
	
	LOG.MLog("Read file header: sign = %s packetsize = %i res = %ix%i vidcodec = %i vidbitrate = %i fps = %i audbitrate = %i audfreq = %i",
			sign,m_nPacketSize,m_nVidWidth,m_nVidHeight,
			m_nVidCodec,m_nVidBitrate,m_nVidFps,m_nAudBitrate,m_nAudSampleRate);
	
	state = FILEPLAY;
	if (b)
		SetStatusbar("Reading from file");
	else
		SetStatusbar("ERROR: Could not interpret file header");
	
	
	b &= (SetupDecRend() == 0);
	
	ToggleConnectButton();
	
	return !b;
}

char decbuf[1048576]; // 1MB
static int filePktSeqNo = 0;

/**
 * @function	DoWorkOnFile()
 * @brief		Does work if in FILEPLAY state
 * @note		It is possible that data will be decoded and rendered after
 * 					the file has completed reading.
 */
void GXLinuxClient::DoWorkOnFile()
{
	//SDL_Delay(20);
	
	if (state != FILEPLAY)
	{
		LOG.MLog("ERROR: Trying to do work on unopened file - exiting!");
		exit(1);
	}
	
	// Let input handler do work
	if (gxl_input.DoWork() == -1)
		CloseClient("Requested by user");
		
	if (file.eof() == 0)
	{
		CGSPktCache::GXPKT pkt;
		
		// Read packet header
		BOOL b = TRUE;
		char ctmp[1024];
		char tmp;
		bzero(ctmp,sizeof(ctmp));
		
		b &= ( file.read(reinterpret_cast<char*>(&pkt.nType),1) != NULL );						//	video-marker(1)
		b &= ( file.read(reinterpret_cast<char*>(&pkt.nTimestamp),sizeof(long long)) != NULL );	// video-sample-timestamp(8)
		b &= ( file.read(reinterpret_cast<char*>(&pkt.nFlags),sizeof(int)) != NULL );			//	video-sample-flags(4)
		b &= ( file.read(reinterpret_cast<char*>(&pkt.nPktBytes),sizeof(int)) != NULL );		// video-sample-size(4)
		
		LOG.DLog("Read header from file: type = %c ts = %lli flag = %i nbytes = %i",
				pkt.nType,pkt.nTimestamp,pkt.nFlags,pkt.nPktBytes);
		
		// Sanity
	    if ((pkt.nPktBytes <= 0) || (pkt.nPktBytes > 1048576))
		{
	    	LOG.DLog("Read packet from file was too large - skipping decode/render");
	    	return;
		}
	
	    // Read the actual data
	    file.read(decbuf,pkt.nPktBytes);
		pkt.pPktData = reinterpret_cast<BYTE*>(decbuf);
	    
		// Inject in packet cache
		m_cache.Put(pkt.nType, pkt.nTimestamp, pkt.nFlags, 0, 1, filePktSeqNo++, pkt.pPktData, pkt.nPktBytes);
	}
	else
		SetStatusbar("Reached end of file");
	
	// Retrieve packet from cache
	CGSPktCache::GXPKT* pPktCC = m_cache.Get();
	if (pPktCC != NULL)
			DoDecQueue(pPktCC);
			
	gxl_renderer.DoWork();
}

//////////////////////////////////////////////////////////
///////////////// DEPRECATED FUNCTIONS ///////////////////
//////////////////////////////////////////////////////////

#if 0
/**
 * @function	ReceiveVideoPacket()
 * @brief		Reads a video packet from the socket and puts it into the packet cache.
 * @returns		-1 if error, 0 if no packet, 1 if something was received
 * @note		flags should be MSG_DONTWAIT to make this non blocking.
 */
int GXLinuxClient::ReceiveVideoPacket(int flags)
{
	if ( state != CONNECTED )
	{
		// TODO: Handle better
		LOG.DLog("ERROR - Trying to receive a packet while not connected to a server");
		return -1;
	}

    // Check socket
    if (m_sock.m_sd == (SOCKET) NULL)
        return -1;
    
    // Get bytes waiting on socket
    if (DoListen() <= 0)
    		return 0;
    
	// Clearing buffers
	bzero(packetBuf,MAX_DATA_PACKET);
	
	// Reading data off buffer
	// TODO: Add check on who the sender is...
	int bytes = recvfrom(m_sock.m_sd, packetBuf, MAX_DATA_PACKET, 0, NULL, NULL);
    
    // Check headers
    char packet_type = *packetBuf;
    
    // Extract the timestamp (seems to be working..?)
    char* tmpChar1 = (char*) calloc(8, sizeof(char));
    strncpy(tmpChar1,packetBuf+1,8);
    int64_t packet_timestamp = *(reinterpret_cast <int64_t*> (tmpChar1));
    free(tmpChar1); 

    // More headers
    int packet_flags = *(packetBuf+9);
    int packet_i = *(packetBuf+13);
    int packet_no = *(packetBuf+17);
    int packet_frameSize = bytes - HEADER_SIZE;
    
    LOG.DLog("Read packet: type=%c ts=%lld flags=%i i=%i no=%i",packet_type,packet_timestamp,packet_flags,packet_i,packet_no);
    
    // Put in packet cache
    if (PktCache.Put(packet_type,packet_timestamp,packet_flags,packet_i,packet_no,(unsigned char*)(packetBuf+21),packet_frameSize) == 0)
    {
    	LOG.MLog("PktCache failed!");
    	return -1;
    }

	return 1;
}

#ifdef USEFFMPEGLIB
    m_decode.SetAudioBitsPerSample(16);
    m_decode.SetAudioChannels(2);
    m_decode.SetAudioSampleRate(44100);
    m_decode.SetAudioCodec(CODEC_ID_WMAV2);
    m_decode.SetVideoWidth(m_nVidWidth);
    m_decode.SetVideoHeight(m_nVidHeight);
    m_decode.SetVideoFps(m_nVidFps);
    //m_decode.SetVideoCodec(m_nVidCodec);
    m_decode.SetVideoCodec(CODEC_ID_WMV2);
    m_decode.RequestVideoFormat(PIX_FMT_RGB32); // FIXME! Should match screen depth.
    if (!m_decode.Create())
    {
    	LOG.MLog("Could not create video decoder");
        return -1;
    }
#endif // USEFFMPEGLIB
    
#ifdef USEFFMPEGLIB
			if (gxl_renderer.RenderFrame(renderFrame) == -1)
#endif
				
// Decode video frame
#ifdef USEFFMPEGLIB
        AVPacket pkt;
        unsigned char* vBuf = NULL;
        int vBufLen = 0;
        pkt.data = (uint8_t*)av_malloc(pPktCC->nPktBytes); // hmm!
        memcpy(pkt.data, pPktCC->pPktData, pPktCC->nPktBytes);
        pkt.size = pPktCC->nPktBytes;
        m_decode.DecodeVideo(&pkt);
        m_decode.GetFrame(&vBuf, &vBufLen);
        av_free(pkt.data);
        
        AVFrame* renderFrame = avcodec_alloc_frame(); 
        avpicture_fill((AVPicture *)renderFrame, vBuf, PIX_FMT_YUV420P, m_nVidWidth, m_nVidHeight);
#endif
        
    	
#ifdef USEFFMPEGLIB
	CGXFFmpegFrameDecoder m_decode;
#endif // USEFFMPEGLIB

/**
 * @function	InitClient()
 * @brief		Sets up the socket information.
 * @returns		-1 on error, 0 if ok
 * @note		DEPRECATED
 */
int GXLinuxClient::InitClient(UserInfo userInfo)
{
	SetUserInfo(userInfo);
	
	si.packetsReceived = 0;
	si.bytesReceived = 0;
	
	packetBuf = (char*) calloc(MAX_DATA_PACKET,sizeof(char));
	decodedFrameSize = 0;
	
	if ( packetBuf == NULL )
	{
		LOG.MLog("Failed to allocate packet/frame cache");
		return -1;
	}
	
	si.socketDescriptor = socket(AF_INET, SOCK_DGRAM, IPPROTO_UDP);
	
	if (si.socketDescriptor == -1)
	{
		LOG.MLog("Failed to create socket");
		return -1;
	}
	
    memset(&si.serverAddress, 0, sizeof(si.serverAddress));
    si.serverAddress.sin_family = AF_INET;
    si.serverAddress.sin_addr.s_addr = Netstuff::ResolveAnyAddress(ui.server);
    si.serverAddress.sin_port = htons(ui.port);
    
    if (&si.serverAddress.sin_addr.s_addr == NULL)
    {
    	LOG.MLog("Failed to create address");
    	CloseClient("Failed to create address");
    	return -1;
    }
    
    return 0;
}

/**
 * @function	PrintStats()
 * @brief		Prints the decoder stats
 * 
 * @note		DEPRECATED
 */
void GXLinuxClient::PrintStats()
{
	return;
	
	printf("\n[GXLinuxClient] *** STATISTICS ***\n\n");
	
	printf("Packets:\t %i\n", si.packetsReceived);
	printf("Bytes:\t\t %i\n", si.bytesReceived);
	printf("\n");
	printf("Time (decode):\t %li ms over %i intervals\n", stats.sw_decode.totalTime/1000.0, stats.sw_decode.GetIntervals() );
	printf("Time (render):\t %li ms over %i intervals\n", stats.sw_render.totalTime/1000.0, stats.sw_render.GetIntervals() );
	printf("Time (total):\t %li ms over %i intervals\n", stats.sw_total.totalTime/1000.0, stats.sw_total.GetIntervals() );
	printf("\n");
	if ( stats.sw_decode.GetIntervals() == stats.sw_render.GetIntervals() )
		printf("No errors detected\nUsed %li ms per video frame on average\n",
				(stats.sw_decode.GetTotalTime()/1000.0 + stats.sw_render.GetTotalTime()/1000.0) / stats.sw_decode.GetIntervals() );
	
	printf("\n[GXLinuxClient] *** /STATISTICS ***\n\n");	
}


/**
 * @function	DoConnectInputTest()
 * @brief		Sets up clientLGV and initiates an input test without using renderer/decoder
 * 
 * @note		DEPRECATED
 */
int GXLinuxClient::DoConnectInputTest()
{
	// Send connect that will bounce, but we ignore it
	DoSendConnectVideo(ui.server, ui.port);
	
	if (gxl_input.InitInputHandlerGXSYS(ui.server,ui.port,NULL,&clientLGV) == -1)
	{
		LOG.MLog("Failed to init the input handler");
		this->CloseClient("Initializing the input handler failed");
		return -1;
	}
	
	while (true)
	{
		gxl_input.DoWork();
	}
	
	return 0;    
}

/**
 * @function	DoConnectLiveStart
 * @brief		Connects the application to LiveStart server.
 * 
 * @returns		-1 if error, 0 if ok
 * 
 * @NOTE		DEPRECATED: Use clientLGV
 * 
 * @TODO		Deal with return message (error, etc).
 */
int GXLinuxClient::DoConnectLiveStart()
{
	int bytes = 0; 
	char datagram[MAX_CONTROL_PACKET];
	
	// CONNECT 2: Connect to LiveStart
	LOG.MLog("[GXLinuxClient] Connecting to LiveStart: %s\n",inet_ntoa(si.serverAddress.sin_addr));
		
	bzero(datagram,MAX_CONTROL_PACKET);
	// TODO: Retrieve GameID for games
	// GameID: Must be '0' for test video
	int gameid = 0;
	sprintf(datagram,"CONNECT GAME\nVERSION 5.0.0.0\n%s\n%s\n%i\n\n",ui.username,ui.password,gameid);
	
	LOG.MLog("--- Sending (%i bytes) ---\n%s--- /Send ---\n",strlen(datagram),datagram);
	if ( (bytes = sendto(si.socketDescriptor, datagram, strlen(datagram)+1, 0, (struct sockaddr *)&si.serverAddress,sizeof(si.serverAddress))) <= 0 )
	{
		throw std::runtime_error("Failed to send CONNECT to LiveStart");
		return -1;
	}
	
	// CONNECT 3: Receive reply from LiveStart
	LOG.MLog("Sent connect to LiveStart, awaiting reply");

	// TODO: Verify sender of packet received
	bzero(datagram,MAX_CONTROL_PACKET);
	bytes = recvfrom(si.socketDescriptor, datagram, MAX_CONTROL_PACKET, 0, NULL, NULL);
	LOG.MLog("--- Received (%i bytes) ---\n%s--- /Receive ---\n",bytes,datagram);
	
	// Checking reply - if "OK" then continue
	if ( strncmp(datagram,"OK",2) != 0 )
	{
		LOG.MLog("LiveStart: Error");
		return -1;
	}
	
	// TODO: Retrieve parameters
	
	// success
	return 0;
}

/**
 * @function	DoConnectHUD()
 * @brief		Connects to the game HUD. 
 * 
 * @returns 	-1 if error, 0 if ok
 * 
 * @NOTE		DEPRECATED: Use clientLGV
 * 
 * @TODO		Actually set video params that is sent out.
 * @TODO		Set desired game ID that is sent out.
 * @TODO		Read and store video params that are sent back in response.
 */
int GXLinuxClient::DoConnectHUD()
{
	int bytes = 0; 
	char datagram[MAX_CONTROL_PACKET];
	
	// CONNECT 4: Connect to HUD
	// TODO: new port (13580) should be included in server response, not be hard coded
	// TODO: set socket rcvbuf size (setsockopt)
	si.serverAddress.sin_port = htons(13580);
	
	char videoparams[MAX_CONTROL_PACKET];
	bzero(videoparams,MAX_CONTROL_PACKET);
	int packetSize = 8192;
	int vidCodec = 19; // TODO
	int audCodec = 0;
	
	// This is an ugly hack and must be fixed at some point.
	sprintf(videoparams,"%i\n%i\n%i\n%i\n%i\n%i\n%i\n%i\n%i\n%i\n%i\n%i\n%i\n\n",packetSize,vidCodec,0,0,0,0,0,audCodec,0,0,0,0,0);

	bzero(datagram,MAX_CONTROL_PACKET);
	int gameid = -1; // TODO
	sprintf(datagram,"CONNECT VIDEO\nVERSION 5.0.0.0\n%s\n%s\n%i\n%s",ui.username,ui.password,gameid,videoparams);	
	
	LOG.MLog("--- Sending (%i bytes) ---\n%s--- /Send ---",strlen(datagram),datagram);
	if ( (bytes = sendto(si.socketDescriptor, datagram, strlen(datagram)+1, 0, (struct sockaddr *)&si.serverAddress,sizeof(si.serverAddress))) <= 0 )
	{
		LOG.MLog("Failed to connect to HUD");
		return -1;
	}
	
	// CONNECT 5: Receive reply from HUD
	LOG.MLog("Sent connect to HUD, awaiting reply");

	// TODO: Verify sender of packet received
	bzero(datagram,MAX_CONTROL_PACKET);
	bytes = recvfrom(si.socketDescriptor, datagram, MAX_CONTROL_PACKET, 0, NULL, NULL);
	LOG.MLog("--- Received (%i bytes) ---\n%s--- /Receive ---",bytes,datagram);
	
	if (hsi.ParsePacket(datagram) != 0)
	{
		LOG.MLog("ERROR: Parsing packet failed");
		return -1;
	}
	
	hsi.Print();
		
	// success
	return 0;
}

////// DEPRECATED BITS AND PIECES ///////
#ifndef USEGXSYS
	// Connect to HUD
	if (this->DoConnectHUD() == -1)
	{
		this->CloseClient("Connecting to game HUD failed");
		return -1;
	}
	
	LOG.MLog("Game HUD Connected");
	SetStatusbar("Game HUD OK");
	
	this->SetConnected();
	
	if (gxl_decoder.InitVideoDecoder(*reinterpret_cast<PixelFormat*>(&hsi.video_pix_fmt),
				*reinterpret_cast<CodecID*>(&hsi.video_codecid),
				hsi.video_width,
				hsi.video_height,
				PIX_FMT_YUV420P) == -1)
	{
		LOG.MLog("Failed to init the video decoder");
		this->CloseClient("Initializing the video decoder failed");
		return -1;
	}
#endif // USEGXSYS

#endif // 0

#endif // GXLINUXCLIENT_H
