/**
 * @file		GXLinuxDecoder.h
 * @author 		Ola Mogstad (olamo@stud.ntnu.no)
 * @updated		19/03/08
 * @class		GXLinuxDecoder
 * @brief		Decodes incoming video frames by calling libav functions
 * @TODO		Fix memory leaks
 */

#ifndef GXLINUXDECODER_H
#define GXLINUXDECODER_H

class GXLinuxDecoder
{
public:
	// constructor & destructor
	GXLinuxDecoder() { videoInit = 0; audioInit = 0; }
	~GXLinuxDecoder() { this->CloseDecoder(); }
	
	// member functions
	int InitVideoDecoder(enum PixelFormat inFormat, enum CodecID codec, int w, int h, enum PixelFormat outFormat);
	int InitAudioDecoder(enum CodecID codec_id, int chans, int samplerate);
	int DecodeVideoPacket(const uint8_t* inCache, int inCacheSize, AVFrame* outFrame, int* outFrameSize);
	int DecodeAudioPacket(AVPacket* pkt, char** outBuf, int* outBufLen);
	void CloseDecoder();
	void PrintVideoParams();
	void PrintStats();

	
	// setters & getters
	enum PixelFormat GetVideoOutFormat() { return this->avVideoOutFormat; }
	AVCodecContext* GetCodecContext() { return aCodecCtx; }
	
	// static functions
	static char* CharPixFmt(enum PixelFormat pix_fmt);
	static char* CharCodecID(enum CodecID codec_id);
	
private:
	// data members
	int videoInit;
	int audioInit;
	
	// Video
	AVCodecContext  *pCodecCtx;
	AVCodec         *pCodec;
	AVFrame         *pFrame; 	// TODO: Change name
	AVFrame         *pFrameRGB;	// TODO: Change name
	uint8_t         *buffer;

#ifdef USESWSCALE
	SwsContext *swsCtx;	// Holds all information on formating
#endif
	enum PixelFormat avVideoOutFormat;
	
	// Audio
	AVCodecContext	*aCodecCtx;
	void* m_aSamples;
	int m_aSamplesLen;
	void* aHoldbackBfr;
	int nHoldbackCtr;
	
};

/**
 * @function	InitVideoDecoder()
 * @brief		Inits the video decoder
 * 
 * @returns		-1 if error, 0 if ok
 */
int GXLinuxDecoder::InitVideoDecoder(enum PixelFormat inFormat, enum CodecID codec, int w, int h, enum PixelFormat outFormat)
{	
	if (videoInit)
	{
		LOG.MLog("Video decoder was already initiated");
		return -1;
	}
	
	LOG.MLog("Initiating video decoder with: inFormat=%i codec=%i w=%i h=%i outFormat=%i",inFormat,codec,w,h,outFormat);
	
	avcodec_init();
	av_register_all();
	pCodecCtx = avcodec_alloc_context();
		
	// Find the decoder for the video stream
	pCodec = avcodec_find_decoder(codec);		
	if (pCodec == NULL) 
	{
		LOG.MLog("Unsupported video codec");
	    return -1; // Codec not found
	}
	
	this->avVideoOutFormat = outFormat;
	pCodecCtx->codec_id = codec;
	pCodecCtx->width = w;
	pCodecCtx->height = h;
	pCodecCtx->pix_fmt = inFormat;
	
	// Allocate video frames
	pFrame = avcodec_alloc_frame();
	pFrameRGB = avcodec_alloc_frame();
	if( (pFrameRGB == NULL) || (pFrame == NULL) )
	{
		LOG.MLog("Could not allocate AVFrame structure");
	    return -1;
	}
	
	// Allocate buffer
	buffer = static_cast<uint8_t*>( av_malloc(avpicture_get_size(avVideoOutFormat, pCodecCtx->width, pCodecCtx->height)) );
	
	// Assign appropriate parts of buffer to image planes in pFrameRGB
	avpicture_fill( reinterpret_cast<AVPicture*>( pFrameRGB ), buffer, avVideoOutFormat, pCodecCtx->width, pCodecCtx->height);
	
	// Specifically for WMV2..?
	if (pCodecCtx->codec_id == CODEC_ID_WMV2)
	{
		pCodecCtx->extradata = (uint8_t*)av_malloc(4);
		pCodecCtx->extradata[0] = 0xf0;
		pCodecCtx->extradata[1] = 0xfa;
		pCodecCtx->extradata[2] = 0xf4;
		pCodecCtx->extradata[3] = 0x80;
		pCodecCtx->extradata_size = 4;
	}

	// Open codec
	if (avcodec_open(pCodecCtx, pCodec) < 0)	// This overwrites pCodecCtx->pix_fmt
	{
		LOG.MLog("Unable to open codec");
		return -1; // Could not open codec
	}
	
#ifdef USESWSCALE
	// Set up the SWScaler
	swsCtx = sws_getContext(pCodecCtx->width, pCodecCtx->height, pCodecCtx->pix_fmt,
			pCodecCtx->width, pCodecCtx->height, avVideoOutFormat, 
			SWS_BICUBIC, NULL, NULL, NULL);
	if(swsCtx == NULL)
	{
		LOG.MLog("Could not initialize the SWScaler");
		return -1;
	}
#endif
	
	videoInit = 1;
	
	PrintVideoParams();
	
	return 0;
}

/**
 * @function	InitAudioDecoder()
 * @brief		Inits the audio decoder things
 * 
 * @returns		-1 if error
 */
int GXLinuxDecoder::InitAudioDecoder(enum CodecID codec_id, int chans, int samplerate)
{
	
	if (audioInit)
	{
		LOG.MLog("Trying to initiate an audio decoder that was already initiated");
		return -1;
	}
	
	LOG.MLog("Initiating audio decoder with: codec_id=%i chans=%i samplerate=%i", codec_id, chans, samplerate);
	LOG.MLog("AVCODEC_MAX_AUDIO_FRAME_SIZE: %i",AVCODEC_MAX_AUDIO_FRAME_SIZE);
	LOG.MLog("AUDIO_DECODER_HOLDBACK: %i",AUDIO_DECODER_HOLDBACK);
	
	// Sanity
	if (samplerate == 0)
	{
		LOG.MLog("Cannot initiate audio decoder with samplerate = 0");
		return -1;
	}
	
	aCodecCtx = avcodec_alloc_context();
	m_aSamples = av_malloc(AVCODEC_MAX_AUDIO_FRAME_SIZE);
    m_aSamplesLen = AVCODEC_MAX_AUDIO_FRAME_SIZE;
    aHoldbackBfr = av_malloc(AUDIO_DECODER_HOLDBACK);
    nHoldbackCtr = 0;
	
    AVCodec *aCodec = avcodec_find_decoder(codec_id);
    if (aCodec == NULL) 
    {
    	LOG.MLog("Could not find audio codec");
    	return -1;
    }
    
    aCodecCtx->channels = chans;
    aCodecCtx->sample_rate = samplerate;
    aCodecCtx->codec_id = codec_id;
    aCodecCtx->extradata = static_cast<uint8_t*>(malloc(10));
    
    // Specifically for WMAV2.
    if (codec_id == CODEC_ID_WMAV2)
    {    
    	aCodecCtx->extradata[0] = 0;
    	aCodecCtx->extradata[1] = 0;
    	aCodecCtx->extradata[2] = 0;
    	aCodecCtx->extradata[3] = 0;
    	aCodecCtx->extradata[4] = 1;
    	aCodecCtx->extradata[5] = 0;
    	aCodecCtx->extradata[6] = 0;
    	aCodecCtx->extradata[7] = 0;
    	aCodecCtx->extradata[8] = 0;
    	aCodecCtx->extradata[9] = 0;
    	aCodecCtx->extradata_size = 10;
    }

    if (avcodec_open(aCodecCtx, aCodec) < 0) return FALSE;
	
	audioInit = 1;
	
	return 0;
}

/**
 * @function	CloseDecoder()
 * @brief		Closes the decoder, freeing memory etc
 */
void GXLinuxDecoder::CloseDecoder()
{
	if (videoInit)
	{
		// Free stuff for WMV2
		av_free(pCodecCtx->extradata);

		// Free the RGB image
		av_free(buffer);
		av_free(pFrameRGB);
		  
		// Free the YUV frame
		av_free(pFrame);
		  
		// Close the codec
		avcodec_close(pCodecCtx);
		
		videoInit = 0;
		
	}

	if (audioInit)
	{
		av_free(aHoldbackBfr);
		av_free(aCodecCtx->extradata);
		av_free(m_aSamples);
		avcodec_close(aCodecCtx);
		
		audioInit = 0;
	}
}

/**
 * @function	DecodeVideoPacket()
 * @brief		Uses the ffmpeg decoder to decode a video packet.
 * @returns		-1 if error, 0 if ok
 */
int GXLinuxDecoder::DecodeVideoPacket(const uint8_t* inCache, int inCacheSize, AVFrame* outFrame, int* outFrameSize)
{
	if (!videoInit)
	{
		LOG.MLog("Trying to decode with uninitiated video decoder");
		return -1;
	}
	
	// Create AVPacket instance
	AVPacket packet;
	if ( av_new_packet(&packet,inCacheSize) != 0 )
	{
		LOG.MLog("Failed to create AVPacket");
		return -1;
	}
	// Somewhat ugly hack - improve..?
	if ( av_dup_packet(&packet) != 0 )
	{
		LOG.MLog("Failed to allocate packet");
		return -1;
	}

	// Copy function parameter data to the AVPacket instance
    memcpy(packet.data, inCache, inCacheSize);
    packet.size = inCacheSize;

    int frameFinished;
    int ret = 0;
    BOOL doConvert = FALSE;
    
    // Check if the codec's pixel format matches desired output format
    if (GetVideoOutFormat() == pCodecCtx->pix_fmt)
    {
    	// Decoding the frame from AVPacket instance to parameter AVFrame instance (outFrame)
    	ret = avcodec_decode_video(pCodecCtx, outFrame, &frameFinished, packet.data, packet.size);
    	doConvert = FALSE;
    }
    else
    {
    	ret = avcodec_decode_video(pCodecCtx, pFrame, &frameFinished, packet.data, packet.size);
    	doConvert = TRUE;
    }
    
    // Some sanity checks
    if ( ret < 0 )
	{
		LOG.MLog("Failed to decode video frame!");
		av_free_packet(&packet);
		return -1;
	}
	if(!frameFinished) 
	{		
		LOG.MLog("Tried to decode an incomplete frame!");
		av_free_packet(&packet);
		return -1;
	}
	
	// Do pixel conversion if needed
	if (doConvert)
	{
		LOG.DLog("Decoder pixel formats were different - converting: GetVideoOutFormat() = %i pCodecCtx->pix_fmt %i",
				GetVideoOutFormat(),pCodecCtx->pix_fmt);
		#ifdef USESWSCALE	// Note: Disabled by default
			sws_scale(swsCtx, pFrame->data, 
					pFrame->linesize, 0, pCodecCtx->height, 
					pFrameRGB->data, pFrameRGB->linesize);
		#else
			img_convert((AVPicture *)pFrameRGB, avVideoOutFormat, (AVPicture*)pFrame, 
			        pCodecCtx->pix_fmt, pCodecCtx->width, pCodecCtx->height);
		#endif
			*outFrame = *pFrameRGB;
	}
	else
		LOG.DLog("Decoder pixel formats were the same - not converting!");
			
	av_free_packet(&packet);
		
	return 0;
}

/**
 * @function	DecodeAudioPacket()
 * @brief		Decodes an audio packet
 * 
 * @returns		-1 if error, 0 if decoded successfully, 1 if output
 */
int GXLinuxDecoder::DecodeAudioPacket(AVPacket* pkt, char** outBuf, int* outBufLen)
{
	if (!audioInit)
	{
		LOG.MLog("Trying to decode with uninitiated audio decoder");
		return -1;
	}

	m_aSamplesLen = AVCODEC_MAX_AUDIO_FRAME_SIZE;
    
    int r = avcodec_decode_audio2(aCodecCtx, static_cast<int16_t*>(m_aSamples),
        &m_aSamplesLen, pkt->data, pkt->size);
    
    if (r < 0) 
    {
        LOG.MLog("Could not decode audio packet");
        return -1;
    }
    
    if (AUDIO_DECODER_HOLDBACK <= 0)
    {
    	*outBuf = static_cast<char*>(m_aSamples);
    	*outBufLen = m_aSamplesLen;
    	return 1;
    }
    
    memcpy(static_cast<char*>(aHoldbackBfr)+nHoldbackCtr,static_cast<char*>(m_aSamples),m_aSamplesLen);
    nHoldbackCtr += m_aSamplesLen;
    
    if (nHoldbackCtr >= AUDIO_DECODER_HOLDBACK)
    {
    	LOG.DLog("Holdback: Counter is %i samples - outputting decoded audio of %i samples",nHoldbackCtr,m_aSamplesLen);
	    
    	// Setting output
    	*outBuf = static_cast<char*>(aHoldbackBfr);
	    *outBufLen = m_aSamplesLen;
	    
	    // Moving data
	    nHoldbackCtr -= m_aSamplesLen;
	    memmove(static_cast<char*>(aHoldbackBfr),static_cast<char*>(aHoldbackBfr) + m_aSamplesLen,nHoldbackCtr);	    
	    
	    return 1;
    }
	else
	{
		LOG.DLog("Holdback: Counter is %i",nHoldbackCtr);
	}
    
    return 0;
}

/**
 * @function	PrintVideoParams()
 * @brief		Prints some parameters of the decoder
 */
void GXLinuxDecoder::PrintVideoParams()
{
	return;
	
	if (this->videoInit == 0)
	{
		LOG.MLog("Decoder not initiated. Cannot print parameters");
		return;
	}
	
	printf("\n[GXLinuxDecoder] *** LINUXDECODER ***\n");
	printf("\n");
	printf("Width:\t\t%i\n",pCodecCtx->width);
	printf("Height:\t\t%i\n",pCodecCtx->height);
	printf("Codec:\t\t%i (%s)\n",pCodecCtx->codec_id,CharCodecID(pCodecCtx->codec_id));
	printf("Format (in):\t%i (%s)\n",pCodecCtx->pix_fmt,CharPixFmt(pCodecCtx->pix_fmt));
	printf("Format (out):\t%i (%s)\n",avVideoOutFormat,CharPixFmt(avVideoOutFormat));
	//printf("Format (in):\t%i (%s)\n",swsCtx->srcFormat,CharPixFmt(swsCtx->srcFormat));
	//printf("Format (in):\t%i (%s)\n",swsCtx->dstFormat,CharPixFmt(swsCtx->dstFormat));
	printf("\n");
	
	printf("[GXLinuxDecoder] *** /LINUXDECODER ***\n\n");
}

/**
 * @function	CharPixFmt()
 * @brief		Returns a pixel format as a pointer to a char
 */
char* GXLinuxDecoder::CharPixFmt(enum PixelFormat pix_fmt) 
{
	switch (pix_fmt)
	{
	case PIX_FMT_RGB24:
		return "PIX_FMT_RGB24";
	case PIX_FMT_RGB32:
		return "PIX_FMT_RGB32";
	case PIX_FMT_YUV420P:
		return "PIX_FMT_YUV420P";
	default:
		return "UNKOWN";
	}
}

/**
 * @function	CharCodecID()
 * @brief		Returns a codec id as a pointer to a char
 */
char* GXLinuxDecoder::CharCodecID(enum CodecID codec_id)
{
	switch (codec_id)
	{
	case CODEC_ID_WMV2:
		return "CODEC_ID_WMV2";
	default:
		return "UNKOWN";
	}
}

#endif // GXLINUXDECODER_H
