/**
 * @file		LiveGamesApp.cpp
 * @author 		Ola Mogstad (olamo@stud.ntnu.no)
 * @updated		07/03/08
 * @brief		All GUI functions
 */

#include "common/globals.h"
#include "common/includes.h"
#include "common/structs.h"

using namespace std;

// Including all proprietary classes
#include "GXLinuxRenderer.h"
#include "GXLinuxDecoder.h"
#include "GXLinuxInputHandler.h"
#include "GXLinuxClient.h"

#include "LiveGamesApp.h"

/**
 * @function 	Constructor 
 * @brief		Gets necessary widgets from Glade file and connects signals to functions.
 */
LiveGamesApp::LiveGamesApp(const char* filename) : mainWindow(0)
{
	// Initiate things
	userInfo.infoRetrieved = 0;
	mainLoopRunning = 0;
	serverList = new ServerInfo[30];
	nServers = 0;
	gameList = new GameInfo[30];
	nGames = 0;
	profiles = new char*[10];
	sizes = new char*[10];
	abitrates = new char*[10];
	nProfiles = 0;
	nSizes = 0;
	naBitrates = 0;
	
	// Initiate libcurl (NOT THREAD SAFE)
	curl_global_init(CURL_GLOBAL_ALL);
			
	// GET XML FILE
	try { refXml = Gnome::Glade::Xml::create(filename); }
	catch (...)
	{
		printf("*** ERROR: Failed to retrieve GUI XML file: %s ***\n",GLADE_FILE);
		printf("Problem: GUI files are not where they are supposed to be\n");
		printf("Did you install this application correctly?\n");
		exit(1);
	}
	
    // WINDOW
	refXml->get_widget("window", mainWindow);
    if (!mainWindow)
        throw std::runtime_error("Could not find window in XML file");
    
    // ABOUT DIALOG
    refXml->get_widget("aboutDialog", aboutDialog);
    if (!aboutDialog)
    	throw std::runtime_error("Could not find about dialog in XML file");
    
    // MSG DIALOG
    refXml->get_widget("msgDialog", msgDialog);
    if (!msgDialog)
    	throw std::runtime_error("Could not find message dialog in XML file");
    
	// STATUS BAR
	refXml->get_widget("statusbar", statusbar);
	if (!statusbar)
		throw std::runtime_error("Could not find statusbar in XML file");
	statusbar->push("Offline",0);
	
	// CONNECT BUTTON
	connectButton = 0;
	refXml->get_widget("connect", connectButton);
	if (!connectButton)
    	throw std::runtime_error("Could not find connect button in XML file");
	connectButton->signal_clicked().connect(sigc::mem_fun(*this, &LiveGamesApp::GUI_BtnConnectClicked));
	
	// PING BUTTON
	Gtk::Button* btn = 0;
	refXml->get_widget("ping", btn);
	if (!btn)
    	throw std::runtime_error("Could not find ping button in XML file");
	btn->signal_clicked().connect(sigc::mem_fun(*this, &LiveGamesApp::GUI_BtnPingClicked));
	
	// REFRESH BUTTON
	btn = 0;
	refXml->get_widget("refresh", btn);
	if (!btn)
    	throw std::runtime_error("Could not find refresh button in XML file");
	btn->signal_clicked().connect(sigc::mem_fun(*this, &LiveGamesApp::GUI_BtnRefreshClicked));
	
	Gtk::ImageMenuItem* menuItem = 0;
	refXml->get_widget("developer",menuItem);
	if (!menuItem)
		throw std::runtime_error("Could not find developer menu item in XML file");
	menuItem->signal_activate().connect(sigc::mem_fun(*this, &LiveGamesApp::ToggleDeveloper));
	
	menuItem = 0;
	refXml->get_widget("sdlrender",menuItem);
	if (!menuItem)
		throw std::runtime_error("Could not find renderer menu item in XML file");
	menuItem->signal_activate().connect(sigc::mem_fun(*this, &LiveGamesApp::ToggleRenderer));
	
	menuItem = 0;
	refXml->get_widget("settimer",menuItem);
	if (!menuItem)
		throw std::runtime_error("Could not find set timer item in XML file");
	menuItem->signal_activate().connect(sigc::mem_fun(*this, &LiveGamesApp::SetTimer));

	// Creating GTK timer (only way to bypass block run() function..?)
	int m_timer_number = 0;
	sigc::slot<bool> my_slot = sigc::bind(sigc::mem_fun(*this,&LiveGamesApp::Timeout), m_timer_number);
	sigc::connection conn = Glib::signal_timeout().connect(my_slot, GTK_TIMER);

	// Setting default stuff
	DefaultLists();
	
	// Pops up an about dialog that must be closed before application launches
	aboutDialog->run();
	aboutDialog->hide();
	
	state = IDLE;
}

/** 
 * @name	Destructor 
 * @TODO	Check memory deallocation..?
 */
LiveGamesApp::~LiveGamesApp()
{
	LOG.MLog("Application terminating");
	curl_global_cleanup();
    delete mainWindow;
}

/**
 * @function	Timeout()
 * @brief		Called with frequency according to GTK_TIMER
 */
bool LiveGamesApp::Timeout(int timernumber)
{	
	client.DoWork();

	return true;
}

/** 
 * @function	GUI_RetrieveUserInput()
 * @brief 		Retrieves user input from main application window and stores it in userInfo struct.
 * 
 * @returns		-1 if error, 0 if okay
 */
int LiveGamesApp::GUI_RetrieveUserInput()
{
	// Clear existing user input
	userInfo.Clear();
	
	Gtk::Entry* entry_username;
	refXml->get_widget("username", entry_username);
	sprintf(userInfo.username,"%s",(char*)entry_username->get_text().c_str());

	Gtk::Entry* entry_password = 0;
	refXml->get_widget("password", entry_password);
	sprintf(userInfo.password,"%s",(char*)entry_password->get_text().c_str());
	
	Gtk::ToggleButton* toggle_fullscreen = 0;
	refXml->get_widget("fullscreen", toggle_fullscreen);
	userInfo.fullscreen = toggle_fullscreen->get_active();
	
	Gtk::ToggleButton* toggle_grabinput = 0;
	refXml->get_widget("grabinput", toggle_grabinput);
	userInfo.grabinput = toggle_grabinput->get_active();
	
	char serverstr[USER_INPUT_SIZE];
	memset(serverstr,'\0',sizeof(serverstr));
	Gtk::ComboBoxEntry* combo_server = 0;
	refXml->get_widget("server",combo_server);
	sprintf(serverstr,"%s",(char*)(combo_server->get_entry())->get_text().c_str());
	// Remove any potential whitespaces in front of server stuff
	int i = 0;
	while (i < sizeof(serverstr))
	{
		if (serverstr[i] != ' ')
			break;
		i++;
	}
	memmove(serverstr,serverstr+i,sizeof(serverstr)-i);
		
	Gtk::ComboBoxEntry* combo_game = 0;
	refXml->get_widget("game",combo_game);
	sprintf(userInfo.game,"%s",(char*)(combo_game->get_entry())->get_text().c_str());
	
	// ComboBox: Profile
	Gtk::ComboBox* combo_profile = 0;
	refXml->get_widget("profile", combo_profile);
	sprintf(userInfo.profile,"%s", profiles[ combo_profile->get_active_row_number() ]);
	
	// ComboBox: Size
	Gtk::ComboBox* combo_size = 0;
	refXml->get_widget("size", combo_size);
	sprintf(userInfo.size,"%s", sizes[ combo_size->get_active_row_number() ]);
	
	// ComboBox: Audio bitrate
	Gtk::ComboBox* combo_audiobr = 0;
	refXml->get_widget("abitrate", combo_audiobr);
	sprintf(userInfo.abitrate,"%s", abitrates[ combo_audiobr->get_active_row_number() ]);
	
	// Entry: Packetsize
	Gtk::Entry* entry_packetsize = 0;
	refXml->get_widget("packetsize", entry_packetsize);
	sscanf((char*)entry_packetsize->get_text().c_str(), "%i", &userInfo.packetsize);
	
	// SpinButton: Framerate
	Gtk::SpinButton* spin_framerate = 0;
	refXml->get_widget("framerate", spin_framerate);
	userInfo.framerate = spin_framerate->get_value_as_int();
	
	if ( strlen(serverstr) <= 0 )
	{
		GUI_PrintStatusbar("Insufficient input");
		userInfo.infoRetrieved = 0;
		return -1;
	}
	
	userInfo.infoRetrieved = 1;
	
	if (strncmp(serverstr,"file:",5) == 0)
	{
		GUI_PrintStatusbar("Server URL was of \"file://\" format - will attempt to read from file");
		strcpy(userInfo.server,serverstr);
	}
	else
	{		
		if ( Netstuff::SplitServerPort(serverstr,userInfo.server,&userInfo.port) == 0 )
			userInfo.port = DEFAULT_CONNECT_PORT;
	}

	LOG.MLog("Retrieved user info: %s",userInfo.ToString().c_str());

	return 0;
}

/**
 * @function	StartClient()
 * @brief		Necessary for SDL threads to work as SDL uses old C semantics
 */
/*int StartClient(void *obj)
{
	GXLinuxClient* client = static_cast<GXLinuxClient*>(obj);
	return client->DoConnect();
}*/

/**
 * @function	GUI_BtnConnectClicked
 * @brief		Calls the two connect functions to LiveStart and HUD.
 */
void LiveGamesApp::GUI_BtnConnectClicked(void)
{
	// If not idle, then the button will act as a disconnect button
	if (state != IDLE)
	{
		client.CloseClient("Requested by user");
		return;
	}
	
	// Retrieving user info
	int ret = GUI_RetrieveUserInput();
	if (ret == -1)
		return;

	client.InitClient(userInfo);
	client.SetGuiComponents(statusbar,connectButton);
	client.DoConnect();
}

/**
 * @function	GUI_BtnPingClicked()
 * @brief		Should ping the server
 * 
 * @TODO		Ping the server
 * @TODO		Add test for bandwidth and packetloss
 * 
 */
void LiveGamesApp::GUI_BtnPingClicked(void)
{
	GUI_PrintStatusbar("The functionality you requested has not yet been implemented");
	return;
}

/**
 * @function	GUI_BtnRefreshClicked()
 * @brief		Should retrieve servers and games from geelix server
 * 
 * @TODO		Retreive servers and games from server
 */
void LiveGamesApp::GUI_BtnRefreshClicked(void)
{	
	GUI_PrintStatusbar("The functionality you requested has not yet been implemented");
	return;
	
	GUI_PrintStatusbar("Retrieving server and game data...");
	while (Gtk::Main::events_pending())
		Gtk::Main::iteration();	// Makes sure statusbar is updated before entering blocked http get
	
	char* serverChar = new char[1024];
	char* gamesChar = new char[1024];
	
	// Retrieve data
	if (RetrieveServerGameData(serverChar, gamesChar) == -1)
	{
		GUI_PrintStatusbar("ERROR: Failed to retrieve server and game data");
		return;
	}
	
	// Parse data
	if (ParseServerGameData(serverChar, gamesChar) == -1)
	{
		GUI_PrintStatusbar("ERROR: Expected different format");
		return;
	}
	
	delete [] serverChar;
	delete [] gamesChar;
	
	PrintServerList();
	PrintGameList();
	
	// Update lists
	if (UpdateLists() == -1)
	{
		GUI_PrintStatusbar("ERROR: Lists could not be updated - DEBUG!");
		return;
	}
	
	GUI_PrintStatusbar("Successfully retrieved server and game data");
}

/**
 * @function	DefaultLists()
 * @brief		Resets the various GUI lists to defaults
 */
void LiveGamesApp::DefaultLists()
{
	// Games
	gameList[0] = GameInfo();
	gameList[0].shortname = "running";
	gameList[0].title = "(Running game)";
	
	nGames = 1;
	
	// Servers
	serverList[0] = ServerInfo();
	serverList[0].id = 0;
	serverList[0].URL = "129.241.209.22:80";
	serverList[0].description = "Unknown title (Norway)";
	
	serverList[1] = ServerInfo();
	serverList[1].id = 1;
	serverList[1].URL = "129.241.209.23:80";
	serverList[1].description ="Half-Life 2: Portal (Norway)";
	
	serverList[2] = ServerInfo();
	serverList[2].id = 2;
	serverList[2].URL = "129.241.209.24:80";
	serverList[2].description ="Trackmania Nations (Norway)";
	
	serverList[3] = ServerInfo();
	serverList[3].id = 3;
	serverList[3].URL = "70.137.216.205:80";
	serverList[3].description ="Unknown title (USA)";
	
	nServers = 4;
	
	// Profiles
	profiles[0] = "256 Kbps";
	profiles[1] = "512 Kbps";
	profiles[2] = "768 Kbps";
	profiles[3] = "1000 Kbps";
	profiles[4] = "2000 Kbps";
	profiles[5] = "3000 Kbps";
	nProfiles = 6;
	
	// Sizes
	sizes[0] = "Normal (1:1)";
	sizes[1] = "Half (1:2)";
	sizes[2] = "Quarter (1:4)";
	nSizes = 3;
	
	// Audio bitrates
	abitrates[0] = "64 Kbps";
	abitrates[1] = "128 Kbps";
	abitrates[2] = "None";
	naBitrates = 3;
	
	UpdateLists();
}

/**
 * @function	UpdateLists()
 * @brief		Update all ComboBoxes in the GUI according to lists 
 *
 * @NOTE		Data must have been retrieved and parsed first
 * @NOTE		Will completely overwrite what already exists
 */
int LiveGamesApp::UpdateLists()
{
	LOG.MLog("Updating the GUI lists");
	//PrintServerList();
	//PrintGameList();
	
	ModelColumns m_Columns;
			
	// Update SERVER
	{
		Gtk::ComboBoxEntry* combo_server = 0;
		refXml->get_widget("server",combo_server);
		Glib::RefPtr<Gtk::ListStore> m_serverModel = Gtk::ListStore::create(m_Columns);	
		
		combo_server->set_model(m_serverModel);
			
		int i;
		
		for (i=0; i<nServers; i++)
		{
			Gtk::TreeModel::Row row;
			row = *(m_serverModel->append());
		
			char* tmp = new char[256];
			sprintf(tmp,"%-22s   %-s",
					serverList[i].URL,serverList[i].description);
			Glib::ustring tmp2((const char*)tmp);
			//row[m_Columns.m_text] = tmp2;
			row.set_value(m_Columns.m_text,tmp2);
			LOG.MLog("Added server: %s",tmp2.c_str());
			delete [] tmp;
		}

	}
	
	// Update GAMES
	{
		Gtk::ComboBoxEntry* combo_games = 0;
		refXml->get_widget("game",combo_games);
		Glib::RefPtr<Gtk::ListStore> m_gamesModel = Gtk::ListStore::create(m_Columns);	
		
		combo_games->set_model(m_gamesModel);
			
		int i;
		
		for (i=0; i<nGames; i++)
		{
			Gtk::TreeModel::Row row;
			row = *(m_gamesModel->append());
		
			char* tmp = new char[256];
			sprintf(tmp,"%s",
					gameList[i].title);
			Glib::ustring tmp2((const char*)tmp);
			//row[m_Columns.m_text] = tmp2;
			row.set_value(m_Columns.m_text,tmp2);
			LOG.MLog("Added game: %s",tmp2.c_str());
			delete [] tmp;
		}
		
		combo_games->set_active(0);
	}
	
	// Update PROFILES
	{
		Gtk::ComboBox* combo_profile = 0;
		refXml->get_widget("profile",combo_profile);
		Glib::RefPtr<Gtk::ListStore> m_gamesModel = Gtk::ListStore::create(m_Columns);	
		
		combo_profile->set_model(m_gamesModel);
			
		int i;
		
		for (i=0; i<nProfiles; i++)
		{
			Gtk::TreeModel::Row row;
			row = *(m_gamesModel->append());
		
			char* tmp = new char[256];
			sprintf(tmp,"%s",
					profiles[i]);
			Glib::ustring tmp2((const char*)tmp);
			//row[m_Columns.m_text] = tmp2;
			row.set_value(m_Columns.m_text,tmp2);
			LOG.MLog("Added profile: %s",tmp2.c_str());
			delete [] tmp;
		}
		
		combo_profile->set_active(3);
	}
	
	// Update SIZES
	{
		Gtk::ComboBox* combo_size = 0;
		refXml->get_widget("size",combo_size);
		Glib::RefPtr<Gtk::ListStore> m_gamesModel = Gtk::ListStore::create(m_Columns);	
		
		combo_size->set_model(m_gamesModel);
			
		int i;
		
		for (i=0; i<nSizes; i++)
		{
			Gtk::TreeModel::Row row;
			row = *(m_gamesModel->append());
		
			char* tmp = new char[256];
			sprintf(tmp,"%s",
					sizes[i]);
			Glib::ustring tmp2((const char*)tmp);
			//row[m_Columns.m_text] = tmp2;
			row.set_value(m_Columns.m_text,tmp2);
			LOG.MLog("Added size: %s",tmp2.c_str());
			delete [] tmp;
		}
		
		combo_size->set_active(0);
	}
	
	// Update ABITRATES
	{
		Gtk::ComboBox* combo_abitrates = 0;
		refXml->get_widget("abitrate",combo_abitrates);
		Glib::RefPtr<Gtk::ListStore> m_gamesModel = Gtk::ListStore::create(m_Columns);	
		
		combo_abitrates->set_model(m_gamesModel);
			
		int i;
		
		for (i=0; i<naBitrates; i++)
		{
			Gtk::TreeModel::Row row;
			row = *(m_gamesModel->append());
		
			char* tmp = new char[256];
			sprintf(tmp,"%s",
					abitrates[i]);
			Glib::ustring tmp2((const char*)tmp);
			//row[m_Columns.m_text] = tmp2;
			row.set_value(m_Columns.m_text,tmp2);
			LOG.MLog("Added audio bitrate: %s",tmp2.c_str());
			delete [] tmp;
		}
		
		combo_abitrates->set_active(0);
	}
	
	return 0;
}

/**
 * @function	PrintServerList()
 * @brief		Prints the server list for debug purposes
 */
void LiveGamesApp::PrintServerList()
{
	int i;
	for (i=0; i<nServers; i++)
	{
		LOG.MLog("%i: %s\n",i,serverList[i].ToString().c_str());
	}
}

/**
 * @function	PrintGameList()
 * @brief		Prints the game list for debug purposes
 */
void LiveGamesApp::PrintGameList()
{
	int i;
	for (i=0; i<nGames; i++)
	{
		LOG.MLog("%i: %s\n",i,gameList[i].ToString().c_str());
	}
}

/**
 * @function	ParseServerGameData()
 * @brief		Parses server and game data
 * @note		RetrieveServerGameData() must have been run
 * @returns		-1 on error, 0 on OK
 * @TODO		Fails on blank fields...
 * @TODO		Remove strtok() / strtok_r() as it is not thread safe
 */
int LiveGamesApp::ParseServerGameData(char* serverChar, char* gamesChar)
{
	// Parsing SERVERS
	{
		LOG.MLog("--- Parsing SERVERS ---\n%s\n",serverChar);
		
		int i = nServers;
		char *s_ptr1, *s_ptr2;
		
		char* s_token = new char[1024];
		s_token = strtok_r(serverChar, "\n", &s_ptr1);
		while(s_token != NULL)
		{    
			LOG.MLog("%i: %s\n",i,s_token);					
			serverList[i] = ServerInfo();
			
			char* s_innertoken = strtok_r(s_token, ";", &s_ptr2);
			sscanf(s_innertoken,"%i",&serverList[i].id);
			s_innertoken = strtok_r(NULL,";", &s_ptr2);
			sscanf(s_innertoken,"%i",&serverList[i].status);
			s_innertoken = strtok_r(NULL,";", &s_ptr2);
			sscanf(s_innertoken,"%i",&serverList[i].reserved);
			s_innertoken = strtok_r(NULL,";", &s_ptr2);
			strcpy(serverList[i].URL,s_innertoken);
			s_innertoken = strtok_r(NULL,";", &s_ptr2);
			strcpy(serverList[i].description,s_innertoken);
			
			LOG.MLog("Added server[%i]: id=%i status=%i res=%i url=%s des=%s",
					i,
					serverList[i].id,
					serverList[i].status,
					serverList[i].reserved,
					serverList[i].URL,
					serverList[i].description
					);
	
			nServers++;
		
			i++;
			s_token = strtok_r(NULL, "\n", &s_ptr1);
		}
	}
	
	// Parsing GAMES
	{
		LOG.MLog("--- Parsing GAMES ---\n%s\n",gamesChar);
		
		int i = nGames;
		char *g_ptr1, *g_ptr2, *g_ptr3;
		char* g_token = strtok_r(gamesChar, "\n", &g_ptr1);
		
		while(g_token != NULL)
		{
			LOG.MLog("%i: %s\n",i,g_token);
			gameList[i] = GameInfo();
			
			char* g_innertoken; 
			
			g_innertoken = strtok_r(g_token, ";", &g_ptr2);
			if (g_innertoken != NULL) sscanf(g_innertoken,"%i",&gameList[i].id);
			g_innertoken = strtok_r(NULL,";", &g_ptr2);
			if (g_innertoken != NULL) strcpy(gameList[i].shortname,g_innertoken);
			g_innertoken = strtok_r(NULL,";", &g_ptr2);
			if (g_innertoken != NULL) strcpy(gameList[i].platform,g_innertoken);
			g_innertoken = strtok_r(NULL,";", &g_ptr2);
			if (g_innertoken != NULL) strcpy(gameList[i].genre,g_innertoken);
			g_innertoken = strtok_r(NULL,";", &g_ptr2);
			if (g_innertoken != NULL) strcpy(gameList[i].releasedate,g_innertoken);
			g_innertoken = strtok_r(NULL,";", &g_ptr2);
			if (g_innertoken != NULL) strcpy(gameList[i].renderer,g_innertoken);
			g_innertoken = strtok_r(NULL,";", &g_ptr2);
			if (g_innertoken != NULL) strcpy(gameList[i].title,g_innertoken);
			g_innertoken = strtok_r(NULL,";", &g_ptr2);
			if (g_innertoken != NULL) strcpy(gameList[i].publisher,g_innertoken);
			g_innertoken = strtok_r(NULL,";", &g_ptr2);
			if (g_innertoken != NULL) strcpy(gameList[i].developer,g_innertoken);
			g_innertoken = strtok_r(g_token, ";", &g_ptr2);
			if (g_innertoken != NULL) sscanf(g_innertoken,"%i",&gameList[i].reserved1);
			g_innertoken = strtok_r(NULL,";", &g_ptr2);
			if (g_innertoken != NULL) strcpy(gameList[i].reserved2,g_innertoken);
			
			// Retrieving comma seperated list of servers
			char* commalist = new char;
			g_innertoken = strtok_r(NULL,";", &g_ptr2);
			if (g_innertoken != NULL) 
			{
					strcpy(commalist,g_innertoken);
				
				int j = 0;
				char* innerg_innertoken = strtok_r(commalist, ",", &g_ptr3);
				while (innerg_innertoken != NULL)
				{
					int tmp = -1;
					sscanf(innerg_innertoken,"%i",&tmp);
					gameList[i].servers[j] = tmp;
					j++;
					innerg_innertoken = strtok_r(NULL,",",&g_ptr3);
				}
			}
			
			LOG.MLog("Added game[%i]: title=%s\n",i,gameList[i].title);
			
			nGames++;
			
			i++;
		    g_token = strtok_r(NULL, "\n", &g_ptr1);
		}
	}

	return 0;
}

/**
 * @function	GUI_PrintStatusbar()
 * @brief		Prints a message on the GUI status bar
 */
void LiveGamesApp::GUI_PrintStatusbar(const char* msg)
{
	LOG.MLog("Statusbar: %s",msg);
	statusbar->push(msg,0);
}

struct Buffer
{
	char* buffer;
	size_t size;
};

size_t URL_data(void *buffer, size_t size, size_t nmemb, void *userp)
{	
	if (size != 1)	// Is buffer a pointer to char?
		return -1;
	
	struct Buffer* tmp = (struct Buffer*) userp;
	memcpy(tmp->buffer, (char*)buffer, nmemb);
	tmp->size = nmemb;
	
	return size*nmemb;
}

/**
 * @function	RetrieveServerGameData()
 * @brief		Retrieves server list and game data from web server
 * 
 * @NOTE		Uses libcurl library for http requests and transfer
 * 
 * @returns		-1 on error, 0 if OK
 */
int LiveGamesApp::RetrieveServerGameData(char* serverChar, char* gamesChar)
{
	struct Buffer bfr;
	bfr.buffer = (char*) calloc(MAX_URL_DATA,sizeof(char));
	
	CURL *curl;
	CURLcode res;
	
	if ( (curl = curl_easy_init()) == NULL )
		return -1;
	
	curl_easy_setopt(curl, CURLOPT_WRITEFUNCTION, URL_data);
	curl_easy_setopt(curl, CURLOPT_WRITEDATA, (void*) &bfr);
	curl_easy_setopt(curl, CURLOPT_USERAGENT, "libcurl-agent/1.0");
	
	// Get SERVERS
	curl_easy_setopt(curl, CURLOPT_URL, URL_GETSERVERS);
	if ( (res = curl_easy_perform(curl)) != 0 )
		return -1;
	memcpy(serverChar,bfr.buffer,bfr.size);
	
	LOG.MLog("--- Received ---\n%s\n--- end ---",serverChar);
		
	// Get GAMES
	curl_easy_setopt(curl, CURLOPT_URL, URL_GETGAMES);
	if ( (res = curl_easy_perform(curl)) != 0 )
		return -1;
	memcpy(gamesChar,bfr.buffer,bfr.size);
	
	LOG.MLog("--- Received ---\n%s\n--- end ---",gamesChar);
	
	curl_easy_cleanup(curl);
	free(bfr.buffer);
	
	return 0;
}

/**
 * @function	ToggleRenderer()
 * @brief		Toggles renderer mode on and off
 */
void LiveGamesApp::ToggleRenderer()
{
	if (RENDER_VIDEO == 1)
	{
		RENDER_VIDEO = 0;
		GUI_PrintStatusbar("Renderer OFF");
	}
	else if (RENDER_VIDEO == 0)
	{
		RENDER_VIDEO = 1;
		GUI_PrintStatusbar("Renderer ON");
	}
}

/**
 * @function	ToggleDeveloper()
 * @brief		Toggles the developer setting
 */
void LiveGamesApp::ToggleDeveloper()
{
#if 0
	
	if (DEVELOPER == 1)
	{
		DEVELOPER = 0;
		GUI_PrintStatusbar("Developer mode OFF");
	}
	else if (DEVELOPER == 0)
	{
		DEVELOPER = 1;
		GUI_PrintStatusbar("Developer mode ON");
	}
#else
	GUI_PrintStatusbar("Developer setting is constant and cannot be changed");
#endif // 0
	
}

/**
 * @function	SetTimer()
 * @brief		Sets the timer setting
 */
void LiveGamesApp::SetTimer()
{
	refXml->get_widget("inputDialog",inputDialog);
	if (!inputDialog)
		throw std::runtime_error("Could not find input dialog in XML file");
	
	inputDialog->show();
	
	Gtk::Button* applyButton;
	refXml->get_widget("applyButton",applyButton);
	if (!applyButton)
		throw std::runtime_error("Could not find apply button in XML file");
	applyButton->signal_clicked().connect(sigc::mem_fun(*this, &LiveGamesApp::SetTimerClick));
	
	Gtk::Entry* fieldEntry = 0;
	refXml->get_widget("fieldEntry", fieldEntry);
	char* tmp = (char*) calloc(20,sizeof(char));
	sprintf(tmp,"%i",GTK_TIMER);
	fieldEntry->set_text(tmp);
	free(tmp);
}

/**
 * @function 	SetTimerClick()
 * @brief		Sets GTK_TIMER according to fieldEntry value, if set.
 */
void LiveGamesApp::SetTimerClick()
{
	int value = 0;
	
	Gtk::Entry* fieldEntry = 0;
	refXml->get_widget("fieldEntry", fieldEntry);
	int ret = sscanf(((char*)fieldEntry->get_text().c_str()),"%i",&value);

	if (ret == 1)
		GTK_TIMER = value;
	
	inputDialog->hide();
}

///////////////////////////////////////////////////
/////////////// DEPRECATED FUNCTIONS //////////////
///////////////////////////////////////////////////

#if 0
menuItem = 0;
	refXml->get_widget("testinput",menuItem);
	if (!menuItem)
		throw std::runtime_error("Could not find test input item in XML file");
	menuItem->signal_activate().connect(sigc::mem_fun(*this, &LiveGamesApp::TestInput));

/**
 * @function	TestInput()
 * @brief		Tests input stream to chosen server, without any connection process
 */
void LiveGamesApp::TestInput(void)
{
	GUI_PrintStatusbar("Initiating input test");
	
	// Retrieving user data
	this->GUI_RetrieveUserInput();
	if (userInfo.infoRetrieved == 0)
		return;
		
	SdlVideo video("Geelix Video","Geelix Video");
	video.AddText("Move mouse curser around in this window.",80,200);
	video.AddText("Input events will be sent to chosen server.",80,215);
	video.AddText("Press Shift + F12 to exit.",80,230);
	
	client.InitClient(userInfo);
	client.SetGuiComponents(statusbar,connectButton);
	client.DoConnectInputTest();
	
	//GXLinuxInputHandler input;
	//input.InitInputHandler(userInfo.server,userInfo.port,NULL);
	//input.DoTestLoop();
}
#endif // 0
