/**
 * @file		Logger.h
 * @author 		Ola Mogstad (olamo@stud.ntnu.no)
 * @updated		23/04/08
 * 
 * @class		Logger
 * @brief		A simple class defining functions for logging to file
 * 
 * 				Defines two key public functions:
 * 					- MLog(): For "main" stuff. Writes to both logs
 * 					- DLog(): For "detailed" stuff. Writes only to the detailed log
 * 
 * TODO			Rewrite into global function
 */

#ifndef LOGGER_H
#define LOGGER_H

#include <stdarg.h>
#include <iostream>
#include <fstream>
#include <time.h>

using namespace std;

#define MAIN_LOG		"gx_main.log"
#define DETAILED_LOG	"gx_details.log"
#define GLOBAL_LOG		"gx_log.log"

/**
 * @function	LogMe()
 * @brief		Global log function
 */
void LogMe(char *fmt, ...)
{
	if (DEVELOPER == 0)
		return;
	
	// Open file
	FILE* file = fopen(GLOBAL_LOG, "a");
	if (file == NULL)
	{
		printf("ERROR: Could not create global log\n");
		exit(1);
	}
	
	// Get process time
	/*timeval now;
	gettimeofday(&now,NULL);
	long timeOfDayMs = now.tv_sec*1000 + now.tv_usec/1000;*/

	// Write to file
	va_list argp;
	fprintf(file, "%11li: ", clock());
	va_start(argp, fmt);	
	vfprintf(file, fmt, argp);
	va_end(argp);
	//fprintf(file, "\n");
	
	// Close file
	fclose(file);
}

class Logger
{
public:
	// constructors and destructors
	Logger();
	~Logger();
	
	// public functions
	void DLog(char *fmt, ...);
	void MLog(char *fmt, ...);
	long GetTimestampUS();
	
	// public data members
	FILE* main;
	FILE* details;

private:
	char* GetTimestamp();
	struct timeval start;
	
};

/**
 * @function	Logger()
 * @brief		Constructor that creates a logfile
 */
Logger::Logger()
{
	if (DEVELOPER == 0)
		return;
	
	// Creating log file (ugly hack)
	char* tmp = new char[300];
	sprintf(tmp,"touch %s; touch %s",MAIN_LOG,DETAILED_LOG);
	system(tmp);
	delete [] tmp;
	
	main = fopen(MAIN_LOG, "a");
	details = fopen(DETAILED_LOG, "a");
		
	if ( (main == NULL) || (details == NULL) )
	{
		printf("ERROR: Could not create log files\n");
		exit(1);
	}
	
	// Writing initiating text to log files
	fprintf(main,"/**\n * @filename\t\t%s\n * @created\t\t%s\n * @CLOCKS_PER_SEC\t%li\n */\n\n", 
			MAIN_LOG, GetTimestamp(), CLOCKS_PER_SEC);
	
	fprintf(details,"/**\n * @filename\t\t%s\n * @created\t\t%s\n * @CLOCKS_PER_SEC\t%li\n */\n\n", 
				DETAILED_LOG, GetTimestamp(), CLOCKS_PER_SEC);
	
	// Getting process start time
	gettimeofday(&start,NULL);
}

/**
 * @function	~Logger()
 * @brief		Destructor that closes logfile
 */
Logger::~Logger()
{	
	if (DEVELOPER == 0)
		return;
	
	fprintf(main, "\n/**\n * @closed\t\t%s\n * @runtime\t\t%li\n */\n\n",GetTimestamp(), GetTimestampUS());
	fprintf(details, "\n/**\n * @closed\t\t%s\n * @runtime\t\t%li\n */\n\n",GetTimestamp(), GetTimestampUS());
	fclose(main);
	fclose(details);
}

/**
 * @function	DLog()
 * @brief		Writes to detailed log only
 */
void Logger::DLog(char *fmt, ...)
{
	if (DEVELOPER == 0)
		return;
	
	va_list argp;
	fprintf(details, "%11li: ", GetTimestampUS());
	va_start(argp, fmt);
	vfprintf(details, fmt, argp);
	va_end(argp);
	fprintf(details, "\n");
}

/**
 * @function	MLog()
 * @brief		Writes to main log and detailed log
 */
void Logger::MLog(char *fmt, ...)
{
	if (DEVELOPER == 0)
		return;
	
	va_list argp;
	fprintf(main, "%11li: ", GetTimestampUS());
	va_start(argp, fmt);
	
	vfprintf(main, fmt, argp);
	//DLog(fmt,argp);
	
	//if (DEVELOPER)
		//printf(fmt,argp);
	
	va_end(argp);
	fprintf(main, "\n");
}

/**
 * @function	GetTimestamp()
 * @brief		Returns current data and time as char
 */
char* Logger::GetTimestamp()
{
	time_t rawtime;
	time(&rawtime);
	char* out = ctime(&rawtime);
	
	// Removing potential newline character that is added by ctime()
	if (out[ strlen(out)-1 ] == '\n')
		out[ strlen(out)-1 ] = '\0';
	
	return out;
}

/**
 * @function	GetTimestampUS()
 * @brief		Returns microseconds since process spawn
 */
long Logger::GetTimestampUS()
{
	timeval now;
	gettimeofday(&now,NULL);
	return (now.tv_sec - start.tv_sec)*1000000 + (now.tv_usec - start.tv_usec);
} 

#endif // LOGGER_H
