#ifndef SUPERBUFFER_H
#define SUPERBUFFER_H

class SuperBuffer
{
public:
	SuperBuffer();
	~SuperBuffer();
	int GetSamples(char* buf, int requestSize);
	int PutSamples(const char* buf, int bufSize);
	int InitSuperBuffer();
	
private:
	void MoveData(int steps);
	
	char* 	data;
	int 	end;
	
};

SuperBuffer::SuperBuffer()
{
	data = (char*) calloc(SUPERBUFFER_SIZE,sizeof(char));
	end = 0;
}

SuperBuffer::~SuperBuffer()
{
	free(data);
}

/**
 * @function	MoveData()
 * @brief		Moves the data after having pulled data from front
 */
void SuperBuffer::MoveData(int steps)
{
	if (end <= steps)
	{
		LOG.DLog("WARNING: MoveData() Something appears to be wrong in super buffer - end=%i steps=%i",end,steps);
		end = 0;
		return;
	}
	
	memmove(data,data+steps,end-steps);
	LOG.DLog("MoveData() moved data %i steps: end %i --> %i",steps,end,end-steps);
	end -= steps;
}

/**
 * @function	PutSamples()
 * @brief		Adds samples to the end of the super buffer. Copies the samples, so input buffer can be
 * 					deallocated right after.
 * @returns		-1 if fail, 0 if success
 */
int SuperBuffer::PutSamples(const char* buf, int bufSize)
{
	if (bufSize <= 0)
		return -1;
	
	// Check if sufficient space
	if (bufSize > (SUPERBUFFER_SIZE - end))
	{
		LOG.DLog("WARNING: PutSamples() SUPERBUFFER IS OUT OF SPACE - IGNORING %i SAMPLES",bufSize);
		return -1;
	}
	
	memcpy(data+end,buf,bufSize);
	LOG.DLog("PutSamples() put %i samples: end %i --> %i",bufSize,end,end+bufSize);
	end += bufSize;
	return 0;
}

/**
 * @function	GetSamples()
 * @brief		Pulls requested amount of samples from data. If not available, then it pulls nothing. The output
 * 					buffer must be allocated in advance.
 * @returns		Amount of samples pulled
 * @note		MUST RUN MOVEDATA() IF SOMETHING WAS PULLED!
 */
int SuperBuffer::GetSamples(char* buf, int requestSize)
{
	// Check if requested size can be met
	if (requestSize > end)
	{
		LOG.DLog("WARNING: GetSamples() requested %i samples when only %i are available - ignoring request!",requestSize,end);
		return 0;
	}
	
	// Copy into output buffer
	memcpy(buf,data,requestSize);
	LOG.DLog("GetSamples() pulled %i samples",requestSize);
	
	// Move remaining data forward
	MoveData(requestSize);
	
	return requestSize;
}

#endif // ! SUPERBUFFER_H
