/**
 * @file		GXLinuxInputHandler.h
 * @author 		Ola Mogstad (olamo@stud.ntnu.no)
 * @updated		28/02/08
 *
 * @class		GXLinuxInputHandler
 * @brief		Handles capturing of input data in an SDL window and sending input stream to a server.
 */

#pragma once
#ifndef GXLINUXINPUTHANDLER_H
#define GXLINUXINPUTHANDLER_H

class GXLinuxRenderer;

class GXLinuxInputHandler
{
public:
	// constructor & destructor
	GXLinuxInputHandler() {}
	~GXLinuxInputHandler() {}

	// member functions
	void CloseInputHandler();
	int DoWork();
	int InitInputHandlerGXSYS(const char* address, int port, GXLinuxRenderer* renderer, CGXClientLGV* cLGV);	
	
private:
	// member functions
	char* PrintState(SDL_KeyboardEvent *key);
	char* PrintModifiers(SDL_KeyboardEvent *key);
	char* PrintKey(SDL_KeyboardEvent *key);
	int DoInput(SDL_Event* event);
	short MapSdlKeyboardToWindowsVK(SDL_KeyboardEvent* event);
	short MapSdlMouseToWindowsVK(SDL_MouseButtonEvent *event);
	
	// deprecated
	attribute_deprecated int DoTestInput(SDL_Event* event);
	attribute_deprecated int DoTestLoop();
	
	// data members
	int initiated;
	int TESTMODE;
	//struct SocketInfo si;
	char* buffer;
	GXLinuxRenderer* gxl_renderer;		// Needed to set full screen, for instance
	CGXClientLGV* clientLGV;
	
	// input specifics
	int seqno;
};

/**
 * @function	InitInputHandler()
 * @brief		Initiates the input handler
 */
int GXLinuxInputHandler::InitInputHandlerGXSYS(const char* address, int port, GXLinuxRenderer* renderer, CGXClientLGV* cLGV)
{	
	if (!SDL_WasInit(SDL_INIT_VIDEO))
	{
		LOG.MLog("Error initiating input handler - SDL was not initiated");
		return -1;
	}
	
	
	if (initiated == 1)
	{
		LOG.MLog("Input handler was already initiated");
		CloseInputHandler();
		// TODO: Handle in a different manner?
	}
	
	clientLGV = cLGV;
	
	if (renderer == NULL)
	{
		LOG.MLog("Renderer was NULL - entering TESTMODE");
		TESTMODE = 1;
	}
	else
		TESTMODE = 0;

	LOG.MLog("Inputhandler initiated with destination address %s:%i",address,port);
	
	gxl_renderer = renderer;
	seqno = 0;
	
	initiated = 1;
	return 0;
}

/**
 * @function	CloseInputHandler()
 * @brief		Closes this handler
 */
void GXLinuxInputHandler::CloseInputHandler()
{
	if (initiated == 1)
	{
		clientLGV = 0;
		gxl_renderer = 0;
		initiated = 0;
	}
}

/**
 * @function	DoWork()
 * @brief		Calls DoInput() on an amount of input events according to LIMITINPUTPUSH
 * @returns		-1 if SDL_QUIT event, 0 otherwise (normal)
 */
int GXLinuxInputHandler::DoWork()
{
	SDL_Event event;
	
	if (LIMITINPUTPUSH <= 0)
	{
		// Exhaust the entire event queue
		if (SDL_PollEvent(&event))
		{
			if (event.type == SDL_QUIT)
				return -1;
			
			DoInput(&event);
			
			// This will cause the function to exhaust the input queue
			return DoWork();
		}
	}
	else
	{
		int extractedEvents = 0;
		
		// Extract a limited amount of input events from queue
		while(extractedEvents <= LIMITINPUTPUSH)
		{
			if (SDL_PollEvent(&event))
			{
				if (event.type == SDL_QUIT)
					return -1;
				
				DoInput(&event);
				extractedEvents++;
			}
			else
				return 0;
		}
	}
	
	return 0;
}

/**
 * @function	DoInput()
 * @brief		Checks if there is an input event, and handles it if there is
 * @returns		-1 if error or exit, 0 if ok
 */
int GXLinuxInputHandler::DoInput(SDL_Event* event)
{
	if ( (event == NULL) || (this->initiated == 0) || (state == FILEPLAY))
		return -1;
	
	int bytes_sent, buffer_length;
	short nDevice = -1, nVirtKey = -1, nData1 = -1, nData2 = -1;
	
	switch(event->type)
		{
			/////////////////////
			// KEYBOARD STUFF
			/////////////////////
	
			case SDL_KEYDOWN:
				
				// Some special functions if SHIFT key
				if (!TESTMODE)
				if (event->key.keysym.mod & (KMOD_LCTRL | KMOD_RCTRL))
				{
					// Have shift+f11 toggle fullscreen
					if (event->key.keysym.sym == SDLK_f)
						gxl_renderer->ToggleFullscreen();
					// Have shift+f12 toggle grab input
					else if (event->key.keysym.sym == SDLK_g)
						gxl_renderer->ToggleGrabInput();
				}
				
			case SDL_KEYUP:
							
				// Set nDevice according to protocol
				nDevice = 1;
				
				// Set nData1 according to protocol
				if (event->type == SDL_KEYDOWN)
					nData1 = 0x80;
				else if (event->type == SDL_KEYUP)
					nData1 = 0x00;
				
				// Retrieve Windows Virtual Key
				nVirtKey = MapSdlKeyboardToWindowsVK(&event->key);
				
				clientLGV->DoSendInput(nDevice, nVirtKey, nData1, nData2);
				break;
			
			///////////////////
			// MOUSE STUFF
			///////////////////
				
			case SDL_MOUSEBUTTONDOWN:
			case SDL_MOUSEBUTTONUP:
			
				nDevice = 2;
				
				// Check if it was a mousewheel event (X handles mwheel as button 4 and 5)
				if (event->button.button == 4 ) 	// mwheelup
				{
					if (event->type == SDL_MOUSEBUTTONUP)	// needed or mwheel will be sent twice
						break;
					
					nVirtKey = 0x01;
					nData1 = 120;		
				}
				else if (event->button.button == 5)	// mwheeldown
				{
					if (event->type == SDL_MOUSEBUTTONUP)	// needed or mwheel will be sent twice
						break;
					
					nVirtKey = 0x01;
					nData1 = -120;		
				}
				else								// other button
				{
					if (event->type == SDL_MOUSEBUTTONDOWN)
						nData1 = 0x80;
					else if (event->type = SDL_MOUSEBUTTONUP)
						nData1 = 0x00;
					
					nVirtKey = MapSdlMouseToWindowsVK(&event->button);
				}
				
				clientLGV->DoSendInput(nDevice, nVirtKey, nData1, nData2);				
				break;
			
			case SDL_MOUSEMOTION:
				
				// Check if mouse and keyboard are grabbed
				if (!TESTMODE)
				if (gxl_renderer->GetGrabInput() == 1)
				{
					// Check if this was the event that moved the mouse to the center -> IGNORE
					if ( (event->motion.x == gxl_renderer->GetWidth()/2) && (event->motion.y == gxl_renderer->GetHeight()/2) )
						return 0;
					
					// Set mouse to center of screen
					SDL_WarpMouse(gxl_renderer->GetWidth()/2, gxl_renderer->GetHeight()/2);
				}
				
				nDevice = 2;
				nVirtKey = 0x00;
				nData1 = event->motion.xrel;
				nData2 = event->motion.yrel;
				
				clientLGV->DoSendInput(nDevice, nVirtKey, nData1, nData2);
				break;

			default:
				break;
		}
	
	return 0;
}

char* GXLinuxInputHandler::PrintState(SDL_KeyboardEvent *key)
{
	if (key->type == SDL_KEYUP)
		return "R";
	else
		return "P";
	
}

char* GXLinuxInputHandler::PrintModifiers(SDL_KeyboardEvent *key)
{
	SDLMod modifier = key->keysym.mod;
	if( modifier & KMOD_NUM ) return "NUMLOCK ";
	if( modifier & KMOD_CAPS ) return "CAPSLOCK ";
	if( modifier & KMOD_MODE ) return "MODE ";
	if( modifier & KMOD_LCTRL ) return "LCTRL ";
	if( modifier & KMOD_RCTRL ) return "RCTRL ";
	if( modifier & KMOD_LSHIFT ) return "LSHIFT ";
	if( modifier & KMOD_RSHIFT ) return "RSHIFT ";
	if( modifier & KMOD_LALT ) return "LALT ";
	if( modifier & KMOD_RALT ) return "RALT ";
	if( modifier & KMOD_LMETA ) return "LMETA ";
	if( modifier & KMOD_RMETA ) return "RMETA ";
	return "";
}

char* GXLinuxInputHandler::PrintKey(SDL_KeyboardEvent *key)
{
	return SDL_GetKeyName(key->keysym.sym);
}

/**
 * @function	MapSdlMouseToWindowsVK()
 * @brief		Maps and SDL mousebutton event to a value according to protocol
 * 
 * @NOTE		DirectInput numerates the buttons from B0 to B7, while linux uses
 * 				B1 and up.
 * 
 * @TODO		Validate that this is correct?!
 */
short GXLinuxInputHandler::MapSdlMouseToWindowsVK(SDL_MouseButtonEvent *event)
{
	// sanity check
	if ( (event->type != SDL_MOUSEBUTTONUP) && (event->type != SDL_MOUSEBUTTONDOWN) )
		return 0x00;
	
	switch (event->button)
	{
	case 1:		return 0x80;
	case 2:		return 0x81;
	case 3:		return 0x82;
	case 4:		return 0x83;
	case 5:		return 0x84;
	case 6:		return 0x85;
	case 7:		return 0x86;
	case 8:		return 0x87;
	}
	
	return 0x00;
}

/**
 * @function	MapSdlKeyboardToWindowsVK()
 * @brief		Takes a keyboardevent and converts it to a short with value according to Windows VK
 * 
 * @TODO		Complete the mappings for keys we don't know what are
 */
short GXLinuxInputHandler::MapSdlKeyboardToWindowsVK(SDL_KeyboardEvent* event)
{
	// Check if this actually was a keybord event
	if ( (event->type != SDL_KEYUP) && (event->type != SDL_KEYDOWN) )
		return 0x00;
	
#ifdef LINUX_ONLY // Deprecated? Only use SDL keymap instead of scancode?
	Uint8 scancode = event->keysym.scancode;
	switch (scancode)
	{
	case 0x9:	return 0x1B;            // escape
	case 0xa:	return 0x31;            // 1
	case 0xb:	return 0x32;            // 2
	case 0xc:	return 0x33;            // 3
	case 0xd:	return 0x34;            // 4
	case 0xe:	return 0x35;            // 5
	case 0xf:	return 0x36;            // 6
	case 0x10:	return 0x37;            // 7
	case 0x11:	return 0x38;            // 8
	case 0x12:	return 0x39;            // 9
	case 0x13:	return 0x30;            // 0
	case 0x14:	return 0x6B;            // + 		??
	case 0x15:	return 0x6F;            // "\" 		??
	case 0x16:  return 0x08;            // backspace
	case 0x17:	return 0x09;            // tab
	case 0x18:	return 0x51;            // q
	case 0x19:	return 0x57;            // w
	case 0x1a:	return 0x45;            // e
	case 0x1b:	return 0x52;            // r
	case 0x1c:	return 0x54;           // t
	case 0x1d:	return 0x59;            // y
	case 0x1e:	return 0x55;            // u
	case 0x1f:	return 0x49;            // i
	case 0x20:	return 0x4F;            // o
	case 0x21:	return 0x50;            // p
	// TODO: case 0x22            // world 69
	// TODO: case 0x23            // compose
	case 0x24:	return 0x0D;            // return
	case 0x25:	return 0xA2;            // left ctrl
	case 0x26:	return 0x41;            // a
	case 0x27:	return 0x53;            // s
	case 0x28:	return 0x44;            // d
	case 0x29:	return 0x46;           	// f
	case 0x2a:	return 0x47;            // g
	case 0x2b:  return 0x48;			// h
	case 0x2c:	return 0x4A;            // j
	case 0x2d:	return 0x4B;            // k
	case 0x2e:	return 0x4C;            // l
	// TODO: case 0x2f            // world 88
	// TODO: case 0x30            // world 70
	// TODO: case 0x31            // unknown key
	case 0x32:	return 0xA0;           // left shift
	// TODO: case 0x33            // '
	case 0x34:	return 0x5A;            // z
	case 0x35:	return 0x58;            // x
	case 0x36:	return 0x43;            // c
	case 0x37:	return 0x56;            // v
	case 0x38:	return 0x42;            // b
	case 0x39:	return 0x4E;            // n
	case 0x3a:	return 0x4D;            // m
	// TODO: case 0x3b            // ,
	// TODO: case 0x3c            // .
	// TODO: case 0x3d            // -
	case 0x3e:	return 0xA1;            // right shift
	case 0x3f:	return 0x6A;            // [*]
	case 0x40:	return 0xA4;            // left alt	?? Menu?
	case 0x41:	return 0x20;            // space
	case 0x42:	return 0x14;            // caps lock
	case 0x43:	return 0x70;            // f1
	case 0x44:	return 0x71;            // f2
	case 0x45:	return 0x72;            // f3
	case 0x46:	return 0x73;            // f4
	case 0x47:	return 0x74;            // f5
	case 0x48:	return 0x75;            // f6
	case 0x49:	return 0x76;            // f7
	case 0x4a:	return 0x77;            // f8
	case 0x4b:	return 0x78;            // f9
	case 0x4c:	return 0x79;            // f10
	
	case 0x52:	return 0x6D;            // [-]
	
	case 0x56:	return 0x6B;            // [+]
	
	// TODO: case 0x5e            		// <
	case 0x5f:	return 0x7A;            // f11
	case 0x60:	return 0x7B;            // f12
	case 0x61:	return 0x24;            // home
	case 0x62:	return 0x26;            // up
	case 0x63:	return 0x22;            // page up
	case 0x64:	return 0x25;            // left
	
	case 0x66:	return 0x27;            // right
	case 0x67:	return 0x23;            // end
	case 0x68:	return 0x28;            // down
	case 0x69:	return 0x21;            // page down
	case 0x6a:	return 0x2D;            // insert
	case 0x6b:	return 0x2E;            // delete
	case 0x6c:	return 0x2B;            // enter
	case 0x6d:	return 0xA3;            // right ctrl
	
	case 0x70:	return 0x6F;            // [/]
	// TODO: case 0x71            // alt gr
	
	// TODO: case 0x73            // left super

	case 0x75:	return 0x12;            // menu
	}
#else // MAC
	switch( event->keysym.sym )
	{
	
	case SDLK_LCTRL:		return 0xA2;            // left ctrl
	case SDLK_RCTRL:		return 0xA3;
	case SDLK_LSHIFT:		return 0xA0;           // left shift
	case SDLK_RSHIFT:		return 0xA1;            // right shift
	case SDLK_ASTERISK:		return 0x6A;            // [*]
	case SDLK_LALT:			return 0xA4;            // left alt	?? Menu?
	case SDLK_SPACE:		return 0x20;            // space
	case SDLK_CAPSLOCK:		return 0x14;            // caps lock
	case SDLK_MINUS:		return 0x6D;            // [-]
	case SDLK_PLUS:			return 0x6B;            // [+]
	case SDLK_HOME:			return 0x24;            // home
	case SDLK_PAGEUP:		return 0x22;            // page up
	case SDLK_PAGEDOWN:		return 0x21;            // page down
	case SDLK_END:			return 0x23;            // end
	case SDLK_INSERT:		return 0x2D;            // insert
	case SDLK_DELETE:		return 0x2E;            // delete
	case SDLK_KP_ENTER:		return 0x2B;            // enter
	case SDLK_SLASH:		return 0x6F;            // [/]
	case SDLK_ESCAPE:		return 0x1b;
	case SDLK_BACKSPACE:	return 0x08;
	case SDLK_TAB:			return 0x09;
	case SDLK_RETURN:		return 0x0d;
	case SDLK_BACKSLASH:	return 0x6f;
	
	case SDLK_UP:			return 0x26;
	case SDLK_DOWN:			return 0x28;
	case SDLK_RIGHT:		return 0x27;
	case SDLK_LEFT:			return 0x25;
	
	case SDLK_F1:			return 0x70;            // f1
	case SDLK_F2:			return 0x71;            // f2
	case SDLK_F3:			return 0x72;            // f3
	case SDLK_F4:			return 0x73;            // f4
	case SDLK_F5:			return 0x74;            // f5
	case SDLK_F6:			return 0x75;            // f6
	case SDLK_F7:			return 0x76;            // f7
	case SDLK_F8:			return 0x77;            // f8
	case SDLK_F9:			return 0x78;            // f9
	case SDLK_F10:			return 0x79;            // f10
	case SDLK_F11:			return 0x7a;
	case SDLK_F12:			return 0x7b;

	case SDLK_a:			return 0x41;
	case SDLK_b:			return 0x42;
	case SDLK_c:			return 0x43;
	case SDLK_d:			return 0x44;
	case SDLK_e:			return 0x45;
	case SDLK_f:			return 0x46;
	case SDLK_g:			return 0x47;
	case SDLK_h:			return 0x48;
	case SDLK_i:			return 0x49;
	case SDLK_j:			return 0x50;
	case SDLK_k:			return 0x51;
	case SDLK_l:			return 0x52;
	case SDLK_m:			return 0x53;
	case SDLK_n:			return 0x54;
	case SDLK_o:			return 0x55;
	case SDLK_p:			return 0x56;
	case SDLK_q:			return 0x57;
	case SDLK_r:			return 0x58;
	case SDLK_s:			return 0x59;
	case SDLK_t:			return 0x60;
	case SDLK_u:			return 0x61;
	case SDLK_v:			return 0x62;
	case SDLK_w:			return 0x63;
	case SDLK_x:			return 0x64;
	case SDLK_y:			return 0x65;
	case SDLK_z:			return 0x66;
	
	case SDLK_1:			return 0x31;            // 1
	case SDLK_2:			return 0x32;            // 2
	case SDLK_3:			return 0x33;            // 3
	case SDLK_4:			return 0x34;            // 4
	case SDLK_5:			return 0x35;            // 5
	case SDLK_6:			return 0x36;            // 6
	case SDLK_7:			return 0x37;            // 7
	case SDLK_8:			return 0x38;            // 8
	case SDLK_9:			return 0x39;            // 9
	case SDLK_0:			return 0x30;            // 0
	}
#endif
	
	return 0x00;
}

////////////////////////////////////////////
////////////// DEPRECATED //////////////////
////////////////////////////////////////////

#if 0
/** 
 * @function	DoTestLoop()
 * @brief		Runs input loop
 * 
 * @NOTE		DEPRECATED
 */
int GXLinuxInputHandler::DoTestLoop()
{
	if ( this->initiated == 0 )
	{
		LOG.MLog("Trying to start loop on uninitiated input handler");
		return -1;
	}
 
	/* enter input event loop */
	SDL_Event event;

	LOG.MLog("Entering input loop");
	
#ifdef SDLINPUTFREQ
	// timer stuff
	float frequency = 0.0;
	long timeQueueTotal = 0, tempDiffTime = 0, oldDiffTime = 0;
	struct timeval nowTime,lastTime;
	queue<long> timeQueue;
#endif
	
	while(initiated)
	{
		//while(SDL_PollEvent(&event))
		if (SDL_WaitEvent(&event) == 0)
			CloseInputHandler();
		
		// reset packet between each iteration
		bzero(buffer,MAX_INPUT_PACKET);
		
#ifdef SDLINPUTFREQ
		// get current time difference
		gettimeofday(&nowTime,NULL);
		tempDiffTime = (nowTime.tv_sec - lastTime.tv_sec)*1000000 + (nowTime.tv_usec - lastTime.tv_usec);
		lastTime = nowTime;

		// add new time difference to queue
		timeQueueTotal += tempDiffTime;
		timeQueue.push(tempDiffTime);
		
		// remove old time difference from queue
		if ( timeQueue.size() > FREQ_COUNT )
		{
			oldDiffTime = timeQueue.front();
			timeQueueTotal -= oldDiffTime;
			timeQueue.pop();
		}
		
		// print updated frequency
		frequency = 1000000.0/((double)timeQueueTotal/(double)timeQueue.size());
		system("clear");
		printf("Frequency: %f Hz\n",frequency);
#endif // SDLINPUTFREQ
		
		if ( DoTestInput(&event) == -1 )
			CloseInputHandler();
		
	}	

	return 0;
}

/**
 * @function	DoTestInput()
 * @brief		Checks if there is an input event, and handles it if there is
 * 
 * @returns		-1 if error or exit, 0 if ok
 * 
 * @NOTE		DEPRECATED
 */
int GXLinuxInputHandler::DoTestInput(SDL_Event* event)
{

	#define TESTINPUT
	
	if ( (event == NULL) || (this->initiated == 0) )
		return -1;
	
	int bytes_sent, buffer_length;
	
	seqno++;
	switch(event->type)
	{
		/*
		* PACKET INFORMATION
		* Delimiter: '\n'
		* Fields: [Type] [SeqNo] [State] {Event} | {MDX MDY}
		*/
		case SDL_KEYDOWN:
			// Quit on shift + f12
			if (event->key.keysym.sym == SDLK_F12)
				return -1;
		case SDL_KEYUP:
			
			// OLD: sprintf(buffer,"LINUX_KB\n0\n%s\n%s",PrintState(&event->key),PrintKey(&event->key));
			if (seqno % 2 == 0)
				sprintf(buffer,"1;%i;0;17",seqno);	// W press
			else
				sprintf(buffer,"1;%i;1;17",seqno);	// W release
			if (DEVELOPER) printf("-------------\n%s\n",buffer);
			buffer_length = strlen(buffer) + 1;
			bytes_sent = sendto(si.socketDescriptor, buffer, buffer_length, 0,(struct sockaddr*) &si.serverAddress, sizeof(struct sockaddr_in) );
			if(bytes_sent < 0)
				LOG.MLog("Failed to send a keyboard input packet");
			break;
		case SDL_MOUSEMOTION:
			
#ifndef TESTINPUT
			// Check if mouse and keyboard are grabbed
			if (gxl_renderer->GetGrabInput() == 1)
			{
				// Check if this was the event that moved the mouse to the center -> IGNORE
				if ( (event->motion.x == gxl_renderer->GetWidth()/2) && (event->motion.y == gxl_renderer->GetHeight()/2) )
					return 0;
				
				// Set mouse to center of screen
				SDL_WarpMouse(gxl_renderer->GetWidth()/2, gxl_renderer->GetHeight()/2);
			}
#endif // ! TESTINPUT
			
			// Create packet with relative position
			sprintf(buffer,"2;%i;3;%i;%i;0",seqno,event->motion.xrel,event->motion.yrel);
			
			// Send packet
			if (DEVELOPER) printf("-------------\n%s\n",buffer);
			buffer_length = strlen(buffer) + 1;
			bytes_sent = sendto(si.socketDescriptor, buffer, buffer_length, 0,(struct sockaddr*) &si.serverAddress, sizeof(struct sockaddr_in) );
			if(bytes_sent < 0)
				LOG.MLog("Failed to send a mouse motion input packet");
			break;
		/*case SDL_QUIT:
			// TODO: Send some quit msg to server
			CloseInputHandler();
			return -1;*/
	}
	
	return 0;
}


#endif

#endif // GXLINUXINPUTHANDLER_H
