/**
 * @file		GXLinuxRenderer.h
 * @author 		Ola Mogstad (olamo@stud.ntnu.no)
 * @updated		25/03/08
 * @class		GXLinuxRenderer
 * @brief		Functions for creating video window with SDL and rendering raw image data (already decoded).
 */

#ifndef GXLINUXRENDERER_H
#define GXLINUXRENDERER_H

static int SDL_BAR = 0;

	class GXLinuxRenderer
	{
	public:
		// constructor & destructor
		GXLinuxRenderer() { this->initiated = 0; }
		~GXLinuxRenderer() {CloseRenderer(); }
		
		// functions
		int InitVideoRenderer(int width, int height, enum PixelFormat src_fmt);
		int InitAudioRenderer(int freq, int chan, int bps);
		void CloseRenderer();
		void EnqueueFrame(AVFrame* frame);
		void EnqueueSamples(char* samples, int len);
		void DoWork();

		// setters & getters
		int ToggleFullscreen();
		void SetGrabInput(int enabled);
		void SetTitlebar(const char* title) { SDL_WM_SetCaption(title, "LiveGames Video"); }
		void ToggleGrabInput();
		int GetGrabInput() { return grabInput; }
		int GetHeight() { return height; }
		int GetWidth() { return width; }
		
		// statics
		static int SaveFramePPM(AVPicture* pic, enum PixelFormat picFmt, int index, int w, int h);
		
		
	private:
		// these should be private as public use might ignore queues
		int RenderFrame(AVFrame* avFrame);
		int RenderSamples();
		
		// functions	
		int AddText(const char* text, int x, int y);
		void apply_surface( int x, int y, SDL_Surface* source, SDL_Surface* destination);
		
		// data member
		int initiated;
		int grabInput;
		int width;
		int height;
		int messageShowFrames;
	
		// Video
		SDL_Surface *screen;		// The window itself
		SDL_Overlay *overlay;		// The YUV overlay where frames are rendered
#ifdef LINUX_ONLY
		TTF_Font* fontVera12;		// The TTF font for writing messages
		SDL_Color colorWhite;		// The color for the font
		SDL_Surface* fontSurface;
#endif
		enum PixelFormat inFormat;
		
#ifdef USESWSCALE
		SwsContext *swsCtx; 	// Needed for conversion to SDL compatible format
#endif

		// Audio
		SDL_AudioSpec	audioSpec;	
		
		// Queues
		AVFrame*		qFrameBfr;	// As of now: only a single buffer for video frames
		int				qFrameNew;	// If 1, new valid frame in queue
		SuperBuffer		audioBuffer;
	};
	
	/**
	 * @function	InitVideoRenderer()
	 * @brief		Initiates the video renderer
	 */
	int GXLinuxRenderer::InitVideoRenderer(int width, int height, enum PixelFormat src_fmt)
	{
		if (this->initiated == 1)
		{
			LOG.MLog("Renderer was already initiated");
			CloseRenderer();
			// TODO: Handle this differently?
		}
		
		
		this->width = width;
		this->height = height;
		inFormat = src_fmt;
		qFrameBfr = avcodec_alloc_frame();
		qFrameNew = 0;
		
		LOG.MLog("[GXLinuxRenderer] Initiating video renderer with: widt=%i height=%i src_fmt=%i",width, height, src_fmt);
		
#ifdef USESWSCALE
		// Set up the SWScaler
		swsCtx = sws_getContext(width, height, src_fmt,
				width, height, PIX_FMT_YUV420P, 
				SWS_BICUBIC, NULL, NULL, NULL);
		if(swsCtx == NULL)
		{
			LOG.MLog("Could not initialize the SWScaler");
			return -1;
		}
#endif
		
		// Init SDL
		if( SDL_Init(SDL_INIT_VIDEO | SDL_INIT_AUDIO | SDL_INIT_TIMER) == -1 ) 
		{
		  fprintf(stderr, "Could not initialize SDL - %s\n", SDL_GetError());
		  return -1;
		}
		
		const SDL_VideoInfo* info;
		
		if ( (info = SDL_GetVideoInfo()) == 0 )
		{
			fprintf(stderr, "SDL: could not query video - aborting\n");
			return -1;
		}
		
		int bpp = info->vfmt->BitsPerPixel;
		
		screen = SDL_SetVideoMode(width, height+SDL_BAR, bpp, SDL_HWSURFACE|SDL_DOUBLEBUF);
		
		if( screen == NULL ) 
		{
			fprintf(stderr, "SDL: could not set SDL to use hardware memory and double buffer - trying gentle mode\n");
			if ( (screen = SDL_SetVideoMode(width, height, bpp, SDL_SWSURFACE|SDL_ANYFORMAT)) == NULL )
			{
				fprintf(stderr, "SDL: could not set SDL video mode - aborting\n");
				return -1;
			}
		}
		
		overlay = SDL_CreateYUVOverlay(width, height, SDL_YV12_OVERLAY, screen);
		
		if ( overlay == NULL )
		{
			fprintf(stderr, "SDL: could not create YUV overlay - aborting\n");
			return -1;
		}
		
		SDL_WM_SetCaption("LiveGames Video", "LiveGames Video");	
	
		initiated = 1;
		grabInput = 0;
		
		// Initing font stuff
#ifdef LINUX_ONLY
		if (TTF_Init() == -1)
		{
			printf("ERROR: Unable to initialize SDL_ttf: %s \n", TTF_GetError());
			LOG.MLog("Could not initiate the renderer");
			return -1;
		}
		fontVera12 = TTF_OpenFont(FONT_VERA12, 12);
		if (fontVera12 == NULL)
		{
			printf("Unable to load font: %s %s \n", FONT_VERA12, TTF_GetError());
			LOG.MLog("Could not initiate the renderer");
			return -1;
		}
		colorWhite.r = 255; colorWhite.g = 255; colorWhite.b = 255;
#endif
		
		// SFont stuff
		/*fontText = new char[256];
		fontImg = IMG_Load(FONT_BITMAP);
		if (fontImg == NULL)
		{
			LOG.MLog("ERROR: Failed to load bitmap font");
			printf("FONT_BITMAP: %s\n",FONT_BITMAP);
			return -1;
		}
		
		SDL_SetAlpha(fontImg, SDL_SRCALPHA|SDL_RLEACCEL, 0);
		
		Font = SFont_InitFont(fontImg);
		if(Font == NULL) 
		{
			LOG.MLog("ERROR: Failed to init sfont");
			return -1;
		}*/
		
		//AddText("TESTING ONE TWO THREE",0,height);
			
		return 0;
	}

	static Uint8 *audio_chunk;
    static Uint32 audio_len;
    static Uint8 *audio_pos;
	char* tmpNewSamples = (char*) malloc(AUDIO_RENDER_BUFFER);
	char* tmp2		= (char*) malloc(AUDIO_RENDER_BUFFER);
   
    /**
     *  C standard
     *  This is called when the audio device is ready for more data
     */
    void fill_audio(void *udata, Uint8 *stream, int len)
    {
    	SuperBuffer* super = (SuperBuffer*) udata;
    	
        // If no data left we should get some
        if ( audio_len < AUDIO_RENDER_BUFFER )
        {
        	/* FIXME: THIS IS A NASTY HACK WITH POSSIBLE MEMORY LEAKS
        	 * SHOULD BE REVISED!!! */
        	
        	bzero(tmpNewSamples,sizeof(char));
        	int nNewSamples = super->GetSamples(tmpNewSamples,AUDIO_RENDER_BUFFER - audio_len);
        	if ( nNewSamples <= 0)
        	{
        		LOG.DLog("WARNING: fill_audio was empty and no data was to be found in super buffer");
        		return;
        	}
        	
        	// First add remainder of audio_pos
        	memmove(tmp2,audio_pos,audio_len);
        	
        	// Then add the new samples
        	memcpy(tmp2+audio_len,tmpNewSamples,nNewSamples);
        	
        	audio_pos = (uint8_t*)tmpNewSamples;
        	LOG.DLog("Added samples: audio_len %i --> %i",audio_len,audio_len+nNewSamples);
        	audio_len += nNewSamples;
        }

        // Mix as much data as possible
        len = ( len > audio_len ? audio_len : len );
        SDL_MixAudio(stream, audio_pos, len, SDL_MIX_MAXVOLUME);
        audio_pos += len;
        audio_len -= len;
        
    }
    
	/**
	 * @function	InitAudioRenderer()
	 * @brief		Inits the audio rendering stuff
	 * 
	 * @returns 	-1 if error
	 */
	int GXLinuxRenderer::InitAudioRenderer(int freq, int chan, int bps)
	{
		if (!SDL_WasInit(SDL_INIT_AUDIO))
		{
			LOG.MLog("Error initiating audio renderer - SDL was not initiated");
			return -1;
		}
		
		LOG.MLog("[GXLinuxRenderer] Initiating audio renderer with: freq=%i chan=%i bps=%i AUDIO_RENDER_BUFFER=%i SUPERBUFFER_SIZE=%i", 
				freq, chan, bps, AUDIO_RENDER_BUFFER, SUPERBUFFER_SIZE);
		
		SDL_AudioSpec wantedSpec;
		wantedSpec.freq = freq;
		wantedSpec.channels = chan;
		wantedSpec.silence = 0;
		
		if (bps == 16)
			wantedSpec.format = AUDIO_S16SYS;
		else if (bps == 8)
			wantedSpec.format = AUDIO_S8;
		else
		{
			LOG.MLog("ERROR: Unknown audio format (bits per sample)");
			return -1;
		}
		
		wantedSpec.samples = AUDIO_RENDER_BUFFER;
		wantedSpec.callback = fill_audio;     
		wantedSpec.userdata = &audioBuffer;

	    // Open audio device
	    if ( SDL_OpenAudio(&wantedSpec, &audioSpec) < 0 ) 
	    {
	        LOG.MLog("ERROR: Could not open audio device: %s", SDL_GetError());
	        return -1;
	    }
	    
	    // This activates audio playback
	    SDL_PauseAudio(0);
	    
	    return 0;		
	}

	/**
	 * @function	CloseRenderer()
	 * @brief		Frees some memory and stuff
	 */
	void GXLinuxRenderer::CloseRenderer()
	{
		if (this->initiated == 1)
		{
			SetGrabInput(0);
			SDL_FreeSurface(screen);
			SDL_FreeYUVOverlay(overlay);
			/*SDL_FreeYUVOverlay(fontOverlay);
			SDL_FreeSurface(fontSurface);
			SFont_FreeFont(Font);
			delete [] fontText;*/
			av_free(qFrameBfr);
			this->initiated = 0;
			SDL_Quit();
		}
	}
	
	/**
	 * @function	DoWork()
	 * @brief		Retrieves data from queues, if any, and renders
	 */
	void GXLinuxRenderer::DoWork()
	{
		/* Nothing is done per iteration for audio as
		 * the SDL callback function is already running
		 * when SDL needs more data */
		
		// VIDEO
		if (qFrameNew == 1)
		{
			RenderFrame(qFrameBfr);
			qFrameNew = 0;
		}
	}
	
	/**
	 * @function	EnqueueFrame()
	 * @brief		Enqueues a frame
	 */
	void GXLinuxRenderer::EnqueueFrame(AVFrame* frame)
	{
		if (qFrameNew == 1)
			LOG.DLog("WARNING: EnqueueFrame() writing over unused frame");
		//memcpy(qFrameBfr,frame,sizeof(frame));
		*qFrameBfr = *frame;
		qFrameNew = 1;
	}


	/**
	 * @function	EnqueueSamples()
	 * @brief		Enqueues audio samples
	 */
	void GXLinuxRenderer::EnqueueSamples(char* samples, int len)
	{
		audioBuffer.PutSamples(samples,len);
	}

//static int frameIndex = 0;
	/**
	 * @function	RenderFrame()
	 * @brief		Reads an RGB AVFrame and renders it on SDL window
	 * 
	 * @TODO		Avoid using any FFMPEG functions
	 */
	int GXLinuxRenderer::RenderFrame(AVFrame* avFrame)
	{			
		if (this->initiated == 0)
		{
			LOG.MLog("Trying to render with uninitiated renderer");
			return -1;
		}
			
		//if ((frameIndex % 100) == 0)
		//	SaveFramePPM((AVPicture*) avFrame, PIX_FMT_YUV420P, frameIndex++, width, height);
		
		// Locking YUV overlay
		SDL_LockYUVOverlay(this->overlay);
		
	    AVPicture pict;
	    pict.data[0] = overlay->pixels[0];
	    pict.data[1] = overlay->pixels[2];
	    pict.data[2] = overlay->pixels[1];
	
	    pict.linesize[0] = overlay->pitches[0];
	    pict.linesize[1] = overlay->pitches[2];
	    pict.linesize[2] = overlay->pitches[1];
		
	    // TODO: if inFormat == PIX_FMT_YUV420P we should not convert!
   		#ifdef USESWSCALE 
				sws_scale(swsCtx, avFrame->data, 
						avFrame->linesize, 0, this->height, 
						pict.data, pict.linesize);
		#else
				img_convert(&pict, PIX_FMT_YUV420P, (AVPicture *)avFrame, 
					        inFormat, width, height);
		#endif

	    SDL_UnlockYUVOverlay(overlay);
	    
	    // Drawing it on the SDL screen
	    SDL_Rect rect;
		rect.x = 0;
		rect.y = 0;
		rect.w = this->width;
		rect.h = this->height;
		SDL_DisplayYUVOverlay(overlay, &rect);
		
		return 0;
	}
		
	/**
	 * @function	SetGrabInput()
	 * @brief		Sets if the mouse should be grabbed by the window or not
	 */
	void GXLinuxRenderer::SetGrabInput(int enable)
	{
		if (enable == 1)
		{
			LOG.MLog("Enabled input grabbing");
			//AddText("Enabled input grabbing - press Shift+F12 to disable",0,height);
			grabInput = 1;
			SDL_ShowCursor(SDL_DISABLE);
			SDL_WM_GrabInput(SDL_GRAB_ON);
		}
		else
		{
			LOG.MLog("Disabled input grabbing");
			//AddText("Disabled input grabbing - press Shift+F12 to enable",0,height);
			grabInput = 0;
			SDL_ShowCursor(SDL_ENABLE);
			SDL_WM_GrabInput(SDL_GRAB_OFF);
		}
	}
	
	/**
	 * @function	ToggleGrabInput()
	 * @brief		Toggles the input grabber
	 */
	void GXLinuxRenderer::ToggleGrabInput()
	{
		if (grabInput)
			SetGrabInput(0);
		else
			SetGrabInput(1);
	}
	
	/**
	 * @function	ToggleFullscreen()
	 * @brief		Toggles fullscreen/windowed mode
	 */
	int GXLinuxRenderer::ToggleFullscreen()
	{
		LOG.MLog("Toggled full screen"); 
		
	#ifdef LINUX ONLY
		return SDL_WM_ToggleFullScreen( this->screen );
	#else // MAC
		screen = SDL_SetVideoMode(screen->w,screen->h,screen->format->BitsPerPixel,
			SDL_HWSURFACE|(screen->flags&SDL_FULLSCREEN?0:SDL_FULLSCREEN));
		overlay = SDL_CreateYUVOverlay(screen->w, screen->h, SDL_YV12_OVERLAY, screen);
		return (screen != NULL);
	#endif
	}
	
	
	/**
	 * @function	SaveFramePPM()
	 * @brief		Saves an AVPicture as a PPM image
	 */
	int GXLinuxRenderer::SaveFramePPM(AVPicture* pic, enum PixelFormat picFmt, int index, int w, int h)
	{
		enum PixelFormat PPM_PIXEL_FORMAT = PIX_FMT_RGB24;
		
		// Convert picture to RGB
		AVFrame* pFrameRGB = avcodec_alloc_frame();
		if(pFrameRGB==NULL)
		  return -1;
		int rgbBytes = avpicture_get_size(PPM_PIXEL_FORMAT, w, h);
		uint8_t* rgbBfr = (uint8_t *) av_malloc(rgbBytes*sizeof(uint8_t));
		avpicture_fill((AVPicture*)pFrameRGB, rgbBfr, PPM_PIXEL_FORMAT, w, h);
		
#ifdef USESWSCALE
		SwsContext* tmpSwsCtx = sws_getContext(w, h, picFmt,
			w, h, PPM_PIXEL_FORMAT, 
			SWS_BICUBIC, NULL, NULL, NULL);
		if(tmpSwsCtx == NULL)
		{
			LOG.MLog("Could not initialize the SWScaler");
			return -1;
		}
		sws_scale(tmpSwsCtx, pic->data, 
			pic->linesize, 0, h, 
			pFrameRGB->data, pFrameRGB->linesize);
#else
		img_convert((AVPicture*)pFrameRGB, PPM_PIXEL_FORMAT, pic, picFmt, w, h);
#endif
		
		FILE *pFile;
		char szFilename[32];
		int  y;
		  
		// Open file
		sprintf(szFilename, "frame%d.ppm", index);
		pFile=fopen(szFilename, "wb");
		if(pFile==NULL)
		{
			LOG.MLog("Could not create file");
			return -1;
		}
		  
		// Write header
		fprintf(pFile, "P6\n%d %d\n255\n", w, h);
	
		// Write pixel data
		for(y=0; y<h; y++)
			fwrite(pFrameRGB->data[0]+y*pFrameRGB->linesize[0], 1, w*3, pFile);
		  
		// Close file
		fclose(pFile);
		av_free(rgbBfr);
		av_free(pFrameRGB);
		
		LOG.MLog("[GXLinuxRenderer] Wrote file: %s",szFilename);
		
		return 0;
	}
	
	/* Adds a char as text on the video window at set coordinates */
	int GXLinuxRenderer::AddText(const char* text, int x, int y)
	{
#ifdef LINUX_ONLY
		SDL_Rect rect;
		rect.x = 0;
		rect.y = height;
		rect.w = width;
		rect.y = SDL_BAR;
		
		SDL_FillRect(screen,&rect,SDL_MapRGB(screen->format,0,0,0));		
		
		/* create the text itself */
		SDL_Surface* message = TTF_RenderText_Solid( fontVera12, text, colorWhite );
		if( message == NULL )
			return -1;
		
		/* add text to screen */
		apply_surface( x, y, message, screen ); 
		SDL_UpdateRect(screen, 0, height, width, SDL_BAR);
		//if( SDL_Flip( screen ) == -1 ) 
			//return -1;
		
		/* clean up */
		SDL_FreeSurface(message);
#endif
		return 0;
	}
	
	/* Applies a surface */
	void GXLinuxRenderer::apply_surface( int x, int y, SDL_Surface* source, SDL_Surface* destination ) 
	{
		SDL_Rect offset;
		offset.x = x;
		offset.y = y;
		SDL_BlitSurface( source, NULL, destination, &offset ); 
	}
	
#if 0
	//////////////////////////////////////////////
	/////////// DEPRECATED ///////////////////////
	//////////////////////////////////////////////
	/**
		 * @function	SetFontMessage()
		 * @brief		Presents a message on display for X frames
		 */
		void GXLinuxRenderer::SetFontMessage(const char* msg, int frames)
		{
			/*fontSurface = TTF_RenderText_Solid( fontVera12, msg, colorWhite );
			if( fontSurface == NULL )
				return;*/
				
			bzero(fontText, sizeof(fontText));
			strcpy(fontText,msg);
			
	printf("SetFontMessage to %s for %i frames\n",msg,frames);
			
			messageShowFrames = frames;
		}
		
		/**
		 * @function	RenderFontMessage()
		 * @brief		Renders the font message if messageShowFrames > 0
		 * 
		 * @note		Not working...
		 */
		void GXLinuxRenderer::RenderFontMessage()
		{
			if (messageShowFrames == 0)
				return;
			
			/*if (fontSurface == NULL)
			{
				LOG.MLog("ERROR: RenderFontMessage failed - font surface was nullpointer");
				return;
			}*/
			
			AddText(fontText,0,height);
			SDL_UpdateRect(screen, 0, height, width, SDL_BAR);
			
		/*	SDL_Overlay* fontOverlay = SDL_CreateYUVOverlay(width, height, SDL_YV12_OVERLAY, fontSurface);
			
			if (fontOverlay == NULL)
			{
				LOG.MLog("ERROR: RenderFontMessage failed - could not convert font surface to YUV overlay");
				return ;
			}
			
	printf("RenderFontMessage: %i frames left\n",messageShowFrames);
			
			// a simple text blit to (0/0)
			SFont_Write (screen, Font, 0,height+SDL_BAR,fontText);

			// Update the screen
			SDL_UpdateRect(screen, 0, height, width, SDL_BAR);
			
			SDL_Rect rect;
			rect.x = 0;
			rect.y = 0;
			rect.w = width;
			rect.h = height;
			SDL_DisplayYUVOverlay(fontOverlay, &rect);
			
			SDL_FreeYUVOverlay(fontOverlay);*/
			
			messageShowFrames--;
		}
#endif // 0
	
	#endif // GXLINUXRENDERER_H
