/**
 * @file		Netstuff.h
 * @author 		Ola Mogstad (olamo@stud.ntnu.no)
 * @updated		07/03/08
 * @brief		Some essential static functions for network usage.
 */

#ifndef NETSTUFF_H
#define NETSTUFF_H

class Netstuff
{
public:
	static in_addr_t ResolveAnyAddress(char* address);
	static int SplitServerPort(char* str, char* server, int* port);
	static int ReceiveUDPStream();
	static int ReceiveRAWStream();
};

#define NETSTUFF_MAX_ADDRESS_LENGTH 100
#define NETSTUFF_MAX_PACKET_SIZE 62000

/**
 * Converts either dot-number IP notation (127.0.0.1)
 * or dns style notation (www.server.com) to an
 * inet_addr.
 */
in_addr_t Netstuff::ResolveAnyAddress(char* address)
{
	in_addr_t outaddr;
	struct hostent *hostInfo;
	
	if ( (outaddr = inet_addr(address)) != INADDR_NONE )
	{
		// Address was in dot-number notation: do nothing
	}
	else if ( (hostInfo = gethostbyname(address)) != NULL )
	{
		// Address was in dns style notation: resolve
		LOG.MLog("[NETSTUFF] %s resolved to %s",hostInfo->h_name, inet_ntoa( *( struct in_addr*)( hostInfo -> h_addr_list[0])));
		outaddr = inet_addr(inet_ntoa(*( struct in_addr*)( hostInfo -> h_addr_list[0])));
	}			 
	else
		// Address was invalid, returning NULL
		outaddr = (in_addr_t) NULL;
		
	return outaddr;
}


/*
 * Splits a string of this format: <serverip>:<port> into
 * a string of server ip and an int for port, respectively.
 * Returns 0 if failed.
 */
int Netstuff::SplitServerPort(char* str, char* server, int* port)
{
	char* tmp = (char*)calloc(NETSTUFF_MAX_ADDRESS_LENGTH,sizeof(char));
	bzero(server,NETSTUFF_MAX_ADDRESS_LENGTH);
	LOG.MLog("[NETSTUFF] Checking split for %s",str);fflush(stdout);
	tmp = strtok(str,":");
	memcpy(server,tmp,strlen(tmp));
	
	if ( (tmp = strtok(NULL,":")) != NULL )
	{
		sscanf(tmp,"%i",port);
		LOG.MLog("[NETSTUFF] Split address into %s and %i",server,*port);
		return 1;
	}
	else
	{
		LOG.MLog("[NETSTUFF] No split on %s",server);
		return 0;
	}
}

/*
 * Receives a stream of UDP packets from somewhere and prints
 * the total amount of bytes received.
 */
int Netstuff::ReceiveUDPStream()
{
	LOG.MLog("[NETSTUFF] *** ReceiveUDPStream started ***");

	int udpSocketDescriptor = socket(PF_INET, SOCK_DGRAM, IPPROTO_UDP);
	
	if ( udpSocketDescriptor == -1 )
	{
		perror("UDP socket failed");
		return -1;
	}
	
	int totalBytes = 0, totalPackets = 0;
	char packet[NETSTUFF_MAX_PACKET_SIZE];
	
	while(totalBytes >= 0)
	{
		bzero(packet,NETSTUFF_MAX_PACKET_SIZE);
		totalBytes += recvfrom(udpSocketDescriptor, packet, NETSTUFF_MAX_PACKET_SIZE, 0, NULL, NULL);
		totalPackets++;
		printf("[NETSTUFF] Received: %i packets %i bytes",totalPackets,totalBytes);
	}
	
	LOG.MLog("[NETSTUFF] *** ReceiveUDPStream stopped ***");
	
	return totalBytes;	
}

/*
 * Receives a stream of RAW packets from somewhere and prints
 * the total amount of bytes received.
 * 
 * @note	Requires root!
 * @TODO	Not working!
 */
int Netstuff::ReceiveRAWStream()
{
	int rawSocketDescriptor = socket(AF_INET, SOCK_RAW, IPPROTO_RAW);
	
	if (rawSocketDescriptor == -1)
	{
		perror("RAW socket failed");
		return -1;
	}
	
	LOG.MLog("[NETSTUFF] *** ReceiveRAWStream started ***");
	
	int bytes = 0, totalBytes = 0, totalPackets = 0;
	char packet[NETSTUFF_MAX_PACKET_SIZE];
	
	while(bytes >= 0)
	{
		bzero(packet,NETSTUFF_MAX_PACKET_SIZE);
		bytes = recvfrom(rawSocketDescriptor, packet, NETSTUFF_MAX_PACKET_SIZE, 0, NULL, NULL);
		totalBytes += bytes;
		totalPackets++;
		printf("[NETSTUFF] Received: %i packets %i bytes",totalPackets,totalBytes); fflush(stdout);
	}
	
	LOG.MLog("[NETSTUFF] *** ReceiveRAWStream stopped ***");
	
	return totalBytes;	
}

#endif // NETSTUFF_H
